/*
 * WItalk2 - italk client for Win32
 * Copyright (C) 1996-1997, TAGA Nayuta <nayuta@is.s.u-tokyo.ac.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

///////////////////////////////////////////////////////////////////////////////
// DialogColor.cxx


#include	<winsock2.h>
#include	<crtdbg.h>
#include	"DialogOrderedList.h"
#include	"WItalk2.h"


class DialogColor : public DialogOrderedList
{
	COLORREF   editing_color;       // ҏW̐F
	COLORREF   focolor_input;       // ̓obt@̕F
	COLORREF   bkcolor_input;       // ̓obt@̔wiF
	COLORREF   bkcolor_log;         // Oobt@̔wiF
	COLORREF   bkcolor_log_public;  // dȊOobt@̔wiF
	COLORREF   bkcolor_log_private; // d̂݃obt@̔wiF
	int        color_right;         // ListBox ̐F\镔̉E[̍W
	LOGFONT    lf;                  // tHg
	BOOL       is_lf_updated;       // lf Cꂽǂ
	HWND       hwnd_fontname;       // tHg\ hwnd
	HWND       hwnd_fontsize;       // tHgTCY\ hwnd

	// tHg̖OƃTCY\
	void set_font(void)
	{
		SetWindowText(hwnd_fontname, lf.lfFaceName);
		HDC hdc = GetDC(HwndPropertySheetOwner());
		SetWindowText(hwnd_fontsize, MulDiv(abs(lf.lfHeight), 72, GetDeviceCaps(hdc, LOGPIXELSY)));
		ReleaseDC(HwndPropertySheetOwner(), hdc);
	}

	void redraw(int id)
	{
		InvalidateRect(Hwnd(id), NULL, TRUE);
		UpdateWindow(Hwnd(id));
	}

public:
	DialogColor(HWND _hwnd)
		: DialogOrderedList(_hwnd),
		  editing_color(RGB(0, 0, 0)),
		  is_lf_updated(FALSE),
		  hwnd_fontname(NULL),
		  hwnd_fontsize(NULL)
	{
	}

	// WM_INITDIALOG
	BOOL wmInitDialog(HWND focus, LPARAM lParam)
	{
		if( lParam ){
			PROPSHEETPAGE* pProp = (PROPSHEETPAGE*)lParam;
			m_pProfileManager = (ProfileManager*)pProp->lParam;
		}
		_ASSERT( m_pProfileManager != NULL );

		registry_name      = "ColorPattern%d";
		new_list_string    = "000000^*$";
		hwnd_edit          = Hwnd(rEditColorPattern);
		hwnd_button_new    = Hwnd(rButtonColorPatternNew);
		hwnd_button_delete = Hwnd(rButtonColorPatternDelete);
		hwnd_button_up     = Hwnd(rButtonColorPatternUp);
		hwnd_button_down   = Hwnd(rButtonColorPatternDown);
		hwnd_list          = Hwnd(rListColorPattern);

		is_lf_updated = FALSE;
		hwnd_fontname = EditOrStatic(Hwnd(rEditFontName));
		hwnd_fontsize = EditOrStatic(Hwnd(rEditFontSize));

		// ListBox ̈ꍀڂ̕w
		HFONT hfont = GetWindowFont(Hwnd(rListColorPattern));
		HDC hdc = GetDC(Hwnd(rListColorPattern));
		hfont = (HFONT)SelectObject(hdc, hfont);
		TEXTMETRIC tm;
		GetTextMetrics(hdc, &tm);
		hfont = (HFONT)SelectObject(hdc, hfont);
		ReleaseDC(Hwnd(rListColorPattern), hdc);
		ListBox_SetItemHeight(Hwnd(rListColorPattern), 0, tm.tmHeight);

		// ListBox ̐F\镔̉E[̍W擾
		RECT rc;
		GetChildWindowRect(Hwnd(rButtonChooseColor), &rc);
		color_right = rc.right;
		GetChildWindowRect(Hwnd(rListColorPattern), &rc);
		color_right -= rc.left;

		m_pProfileManager->Read("FoColorInput", (int &)focolor_input, GetSysColor(COLOR_WINDOWTEXT));
		m_pProfileManager->Read("BkColorInput", (int &)bkcolor_input, GetSysColor(COLOR_WINDOW));
		m_pProfileManager->Read("BkColorLog",	(int &)bkcolor_log, GetSysColor(COLOR_WINDOW));
		m_pProfileManager->Read("BkColorLogPublic", (int &)bkcolor_log_public, GetSysColor(COLOR_WINDOW));
		m_pProfileManager->Read("BkColorLogPrivate", (int &)bkcolor_log_private, GetSysColor(COLOR_WINDOW));

		m_pProfileManager->Read("Font", lf, LoadString(rStringDefaultLogFont));
		set_font();

		return DialogOrderedList::wmInitDialog(focus, lParam);
	}

	// I[i[h[
	BOOL wmDrawItem(int id, DRAWITEMSTRUCT *dis)
	{
		switch (id) {
		case rButtonChooseColor:
			return wmDrawItem_ButtonColor(dis, editing_color);
		case rButtonFoColorInput:
			return wmDrawItem_ButtonColor(dis, focolor_input);
		case rButtonBkColorInput:
			return wmDrawItem_ButtonColor(dis, bkcolor_input);
		case rButtonBkColorLog:
			return wmDrawItem_ButtonColor(dis, bkcolor_log);
		case rButtonBkColorLogPublic:
			return wmDrawItem_ButtonColor(dis, bkcolor_log_public);
		case rButtonBkColorLogPrivate:
			return wmDrawItem_ButtonColor(dis, bkcolor_log_private);
		case rListColorPattern:
			return wmDrawItem_ListColorPattern(dis);
		}
		return TRUE;
	}

	// F\{^`
	BOOL wmDrawItem_ButtonColor(DRAWITEMSTRUCT *dis, COLORREF color)
	{
		RECT rc;
		GetClientRect(dis->hwndItem, &rc);
		if (dis->itemState & ODS_SELECTED)
			lgGUIParts::Draw3DFrameInside(dis->hDC, rc, lgGUIParts::SortOf3DFrame::ButtonDown, COLOR_3DFACE);
		else if (dis->itemState & ODS_DISABLED)
			lgGUIParts::Draw3DFrame(dis->hDC, rc, lgGUIParts::SortOf3DFrame::ButtonUp, COLOR_3DFACE);
		else
			lgGUIParts::Draw3DFrameInside(dis->hDC, rc, lgGUIParts::SortOf3DFrame::ButtonUp, COLOR_3DFACE);

		rc.left   += 2;
		rc.top    += 2;
		rc.bottom -= 2;
		rc.right  -= 2;
		if (dis->itemState & ODS_SELECTED) rc.left++, rc.top++, rc.right++, rc.bottom++;
		HBRUSH hbr = CreateSolidBrush(ItalkView::GetColor(color));
		FillRect(dis->hDC, &rc, hbr);
		DeleteObject(hbr);
		if (dis->itemState & ODS_FOCUS) DrawFocusRect(dis->hDC, &rc);

		return TRUE;
	}

	// ListColorPattern `
	BOOL wmDrawItem_ListColorPattern(DRAWITEMSTRUCT *dis)
	{
		if (0 <= dis->itemID) {
			char buf[BUFSIZE];
			if (LB_ERR != ListBox_GetText(Hwnd(rListColorPattern), dis->itemID, buf)) {
				COLORREF c = ItalkView::ColorPattern::GetColor(buf);
				COLORREF fg, bg;
				if (dis->itemState & ODS_SELECTED) fg = bkcolor_log, bg = c;
				else fg = c, bg = bkcolor_log;
				HBRUSH hbr = CreateSolidBrush(bg);
				FillRect(dis->hDC, &(dis->rcItem), hbr);
				DeleteObject(hbr);

				RECT rc = dis->rcItem;
				rc.left ++;
				rc.top ++;
				rc.right = color_right - 2;
				rc.bottom --;
				hbr = CreateSolidBrush(c);
				FillRect(dis->hDC, &rc, hbr);
				DeleteObject(hbr);

				SetTextColor(dis->hDC, fg);
				SetBkColor(dis->hDC, bg);
				const char *text = ItalkView::ColorPattern::GetPattern(buf);
				rc = dis->rcItem;
				rc.left = color_right;
				DrawText(dis->hDC, text, strlen(text), &rc, DT_LEFT);

				if (dis->itemState & ODS_FOCUS) {
					SetTextColor(dis->hDC, RGB(255, 255, 255));
					DrawFocusRect(dis->hDC, &(dis->rcItem));
				}
			}
		}
		return TRUE;
	}

	// OK ͍XV
	BOOL wmNotify_psnApply(void)
	{
		BOOL r = DialogOrderedList::wmNotify_psnApply();

		m_pProfileManager->Write("FoColorInput"     , focolor_input);
		m_pProfileManager->Write("BkColorInput"     , bkcolor_input);
		m_pProfileManager->Write("BkColorLog"       , bkcolor_log);
		m_pProfileManager->Write("BkColorLogPublic" , bkcolor_log_public);
		m_pProfileManager->Write("BkColorLogPrivate", bkcolor_log_private);
		m_pProfileManager->Write("Font", lf);

		witalk2_ptr->foColorInput = focolor_input; witalk2_ptr->bkColorInput = bkcolor_input;
		witalk2_ptr->bkColorLog = bkcolor_log; witalk2_ptr->bkColorLogPublic = bkcolor_log_public;
		witalk2_ptr->bkColorLogPrivate = bkcolor_log_private;

		SendMessage(HwndPropertySheetOwner(), WM_DIALOGNOTIFY, rDialogColor, is_lf_updated);
		return r;
	}

	// WM_COMMAND
	BOOL wmCommand(int notify_code, int id, HWND hwnd_control)
	{
		command = notify_code;
		switch (id) {
#define c(x) case r##x: return wmCommand_##x()
			c(ButtonChooseColor);
			c(ButtonFoColorInput);
			c(ButtonFoColorReset);
			c(ButtonBkColorInput);
			c(ButtonBkColorLog);
			c(ButtonBkColorLogPublic);
			c(ButtonBkColorLogPrivate);
			c(ButtonBkColorReset);
			c(ButtonChooseFont);
#undef c
		}
		return DialogOrderedList::wmCommand(notify_code, id, hwnd_control);
	}

	void GetEditText(char *buf, int sizeof_buf)
	{
		ItalkView::ColorPattern::SetColor(buf, editing_color);
		GetWindowText(rEditColorPattern, buf + 6, sizeof_buf - 6);
	}

	void SetEditText(char *buf)
	{
		editing_color = ItalkView::ColorPattern::GetColor(buf);
		redraw(rButtonChooseColor);
		SetWindowText(rEditColorPattern, ItalkView::ColorPattern::GetPattern(buf));
	}

	BOOL wmCommand_ButtonChooseColor(void)
	{
		if (command == BN_CLICKED) if (ChooseColor(Hwnd(), editing_color)) {
			command = EN_UPDATE;
			wmCommand_Edit();
			redraw(rButtonChooseColor);
		}
		return TRUE;
	}

	BOOL wmCommand_ButtonFoColorInput(void)
	{
		if (command == BN_CLICKED) if (ChooseColor(Hwnd(), focolor_input)) {
			redraw(rButtonFoColorInput);
			is_lf_updated = TRUE; // ̓obt@邽
			PropertySheetChanged();
		}
		return TRUE;
	}

	BOOL wmCommand_ButtonFoColorReset(void)
	{
		focolor_input = (COLORREF)~COLOR_WINDOWTEXT;
		redraw(rButtonFoColorInput);
		return TRUE;
	}

	BOOL wmCommand_ButtonBkColorInput(void)
	{
		if (command == BN_CLICKED) if (ChooseColor(Hwnd(), bkcolor_input)) {
			redraw(rButtonBkColorInput);
			PropertySheetChanged();
		}
		return TRUE;
	}

	BOOL wmCommand_ButtonBkColorLog(void)
	{
		if (command == BN_CLICKED) if (ChooseColor(Hwnd(), bkcolor_log)) {
			redraw(rButtonBkColorLog);
			redraw(rListColorPattern);
			PropertySheetChanged();
		}
		return TRUE;
	}

	BOOL wmCommand_ButtonBkColorLogPublic(void)
	{
		if (command == BN_CLICKED) if (ChooseColor(Hwnd(), bkcolor_log_public)) {
			redraw(rButtonBkColorLogPublic);
			PropertySheetChanged();
		}
		return TRUE;
	}

	BOOL wmCommand_ButtonBkColorLogPrivate(void)
	{
		if (command == BN_CLICKED) if (ChooseColor(Hwnd(), bkcolor_log_private)) {
			redraw(rButtonBkColorLogPrivate);
			PropertySheetChanged();
		}
		return TRUE;
	}
  
	BOOL wmCommand_ButtonBkColorReset(void)
	{
		bkcolor_input       = GetSysColor(COLOR_WINDOW);
		bkcolor_log         = GetSysColor(COLOR_WINDOW);
		bkcolor_log_public  = GetSysColor(COLOR_WINDOW);
		bkcolor_log_private = GetSysColor(COLOR_WINDOW);
		redraw(rButtonBkColorInput);
		redraw(rButtonBkColorLog);
		redraw(rButtonBkColorLogPublic);
		redraw(rButtonBkColorLogPrivate);
		return TRUE;
	}

	BOOL wmCommand_ButtonChooseFont(void)
	{
		if (command == BN_CLICKED) {
			LOGFONT _lf = lf;
			CHOOSEFONT cf;
			cf.lStructSize = sizeof(CHOOSEFONT);
			cf.hwndOwner = Hwnd();
			cf.lpLogFont = &_lf;
			cf.Flags = CF_FORCEFONTEXIST | CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
			cf.lCustData = NULL;
			cf.lpfnHook = NULL;
			if (ChooseFont(&cf)) {
				is_lf_updated = TRUE;
				lf = _lf;
				set_font();
				PropertySheetChanged();
			}
		}
		return TRUE;
	}
};


BOOL CALLBACK DialogColorDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
#define lgWindowClass DialogColor
	if (!lgwc) switch (message) {
		lgwmInitDialog;
	} else switch (message) {
		lgwmCommand;
		lgwmDrawItem;
	}
	return lgwc->Default(message, wParam, lParam, TRUE);
}
