/*
 * WItalk2 - italk client for Win32
 * Copyright (C) 1996-1997, TAGA Nayuta <nayuta@is.s.u-tokyo.ac.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

///////////////////////////////////////////////////////////////////////////////
// ItalkView.cxx


#include	<winsock2.h>
#include	<crtdbg.h>

#include	"ItalkView.h"
#include	<process.h>

#include	<richedit.h>

#include	"WItalk2.h"

#include	"TrapFocus.h"
#include	"BackLogWindow.h"
#include	"ProfileManager.h"
#include	"DialogURL.h"
#include	"lgString.h"

extern	LOGFONT	gbl_lf;


int	ItalkView::fontheightTwip = 0;
int	ItalkView::lineheightTwip = 0;
int	ItalkView::pxPerInch = 0;


int ItalkView::ColorPattern::ctoi16(char c)
{
	if ('0' <= c && c <= '9') return c - '0';
	if ('a' <= c && c <= 'f') return c - 'a' + 10;
	if ('A' <= c && c <= 'F') return c - 'A' + 10;
	return 0;
}


int ItalkView::ColorPattern::ctoi16(wchar_t c)
{
	if (L'0' <= c && c <= L'9') return c - L'0';
	if (L'a' <= c && c <= L'f') return c - L'a' + 10;
	if (L'A' <= c && c <= L'F') return c - L'A' + 10;
	return 0;
}


void ItalkView::ColorPattern::SetColor(char *buf, COLORREF c)
{
	char tmp = buf[6];
	sprintf(buf, "%02x%02x%02x", GetRValue(c), GetGValue(c), GetBValue(c));
	buf[6] = tmp;
}


void ItalkView::ColorPattern::SetColor(wchar_t *buf, COLORREF c)
{
	wchar_t tmp = buf[6];
	swprintf(buf, L"%02x%02x%02x", GetRValue(c), GetGValue(c), GetBValue(c));
	buf[6] = tmp;
}


COLORREF ItalkView::ColorPattern::GetColor(const char *str)
{
	if (strlen(str) < 6) return RGB(0, 0, 0);
	return RGB(ctoi16(str[0]) * 16 + ctoi16(str[1]),
			   ctoi16(str[2]) * 16 + ctoi16(str[2]),
			   ctoi16(str[4]) * 16 + ctoi16(str[4]));
}


COLORREF ItalkView::ColorPattern::GetColor(const wchar_t *str)
{
	if (wcslen(str) < 6) return RGB(0, 0, 0);
	return RGB(ctoi16(str[0]) * 16 + ctoi16(str[1]),
			   ctoi16(str[2]) * 16 + ctoi16(str[2]),
			   ctoi16(str[4]) * 16 + ctoi16(str[4]));
}


const char *ItalkView::ColorPattern::GetPattern(const char *str)
{
	if (strlen(str) < 6) return str;
	return str + 6;
}


const wchar_t *ItalkView::ColorPattern::GetPattern(const wchar_t *str)
{
	if (wcslen(str) < 6) return str;
	return str + 6;
}


void ItalkView::ColorPattern::Set(const wchar_t *color_pattern)
{
	pattern = wstring(GetPattern(color_pattern));
	color   = GetColor(color_pattern);
	int len = pattern.length();
	if (0 < len && pattern[len - 1] == L'$') // s͕K \r\n ŏI
		pattern[len - 1] = L'\r';
}


// ItalkView


// b`GfBbgRg[Ɍ݂̃tHgݒ肷
void ItalkView::set_current_font(BOOL does_set_color)
{
	color_cf.dwMask = CFM_BOLD | CFM_FACE | CFM_ITALIC | CFM_SIZE | CFM_STRIKEOUT | CFM_UNDERLINE;
	if (does_set_color) color_cf.dwMask |= CFM_COLOR;
	Re.SetCharFormat(&color_cf, SCF_SELECTION);
}

void ItalkView::set_current_font2000(BOOL does_set_color)
{
	color_cf.dwMask = CFM_BOLD | CFM_FACE | CFM_ITALIC | CFM_SIZE | CFM_STRIKEOUT | CFM_UNDERLINE;
	if (does_set_color) color_cf.dwMask |= CFM_COLOR;
	Re.SetCharFormat(&color_cf, SCF_ALL);
	Re.SendMessage(EM_SETLANGOPTIONS, 0, 0);
	gbl_lf.lfCharSet = SHIFTJIS_CHARSET;
	HFONT	hFont = CreateFontIndirect(&gbl_lf);
	Re.SendMessage(WM_SETFONT, (WPARAM)hFont, 0);
}


// b`GfBbgRg[͈̂͂ɐFݒ肷
void ItalkView::set_color(int char_start, int char_end, COLORREF color)
{
	color_cf.dwMask = CFM_COLOR;
	color_cf.crTextColor = color;

	// Ft
	CHARRANGE cr_bak;
	Re.GetSel(&cr_bak);
	Re.SuperSetSel(char_start, char_end);
	Re.SetCharFormat(&color_cf, SCF_SELECTION);
	Re.ReleaseSetSel();
	Re.SetSel(&cr_bak);
}

// b`GfBbgRg[͈̂͂ URL pݒ肷
void ItalkView::set_urlunderline(int char_start, int char_end)
{
	color_cf.dwMask = CFM_COLOR | CFM_UNDERLINE;
	color_cf.dwEffects = CFE_UNDERLINE;
	color_cf.bUnderlineType = CFU_UNDERLINENONE;
	if (DialogURL::url_clicktoopen) {
		color_cf.dwMask |= CFM_LINK; color_cf.dwEffects |= CFE_LINK;
	}
	color_cf.crTextColor = 0x00FF0000;

	CHARRANGE	cr_bak;
	Re.GetSel(&cr_bak);
	Re.SuperSetSel(char_start, char_end+1);
	Re.SetCharFormat(&color_cf, SCF_SELECTION);
	Re.ReleaseSetSel();
	
	color_cf.dwMask = CFM_COLOR;
	
	Re.SetSel(&cr_bak);
}


ItalkView::ItalkView(HWND parent, WItalk2* witalk2_parent, COLORREF* _foColor, COLORREF* _bkColor, DWORD style, DWORD default_style)
					 : Re(CreateWindowEx(WS_EX_CLIENTEDGE, RICHEDIT_CLASS, NULL, style | default_style, 0, 0, 1, 1,
										 parent, NULL, lgWindows::hInst(), NULL))
{
	// 
	color_cf.cbSize = sizeof(CHARFORMAT2);
	witalk2parent = witalk2_parent;
	color_pattern = NULL;
	foColor = _foColor; bkColor = _bkColor;
	scdata = NULL;

	ReadRegistry(FALSE);
	Re.SetTextLimitMaximum();

	lastScrolledLineCount = 0;

	set_current_font2000(TRUE);
	SetLinegap();

	m_bPostQuitMsg = TRUE;
}


ItalkView::~ItalkView()
{
	if( m_bPostQuitMsg ){
		PostMessage(Re.Hwnd(), WM_QUIT, 0, 0);
	}
	if (color_pattern) delete [] color_pattern;
}

void ItalkView::setSubclass(bool inputMode, ReSubClass* _scdata)
{
	scdata = _scdata;
	scdata->oldWndProc = (WNDPROC)GetWindowLongPtr(Re.Hwnd(), GWLP_WNDPROC);
	scdata->inputMode = inputMode;
	SetWindowLongPtr(Re.Hwnd(), GWLP_USERDATA, (LONG_PTR)scdata);
	SetWindowLongPtr(Re.Hwnd(), GWLP_WNDPROC, (LONG_PTR)((inputMode) ? RichEditSubclassInput : RichEditSubclassLog));
}

void ItalkView::setSubclass(bool inputMode)
{
	setSubclass(inputMode, new ReSubClass());
}

void ItalkView::setParentPointer(WItalk2* witalk2_parent){ witalk2parent = witalk2_parent; }

// Waffle 0.2.1a:
//   sԐݒ
void ItalkView::SetLinegapSel()
{
	PARAFORMAT2		pform2;
	//lgRegistry	reg(REGISTRY_ROOT);
	ProfileManager* pProfile = witalk2parent->getProfileManager();
	int		enableLinegap;
	pProfile->Read("EnableLinegap", enableLinegap, false);
	pform2.cbSize = sizeof(PARAFORMAT2); pform2.dwMask = PFM_LINESPACING | PFM_SPACEAFTER;
	Re.SendMessage(EM_GETPARAFORMAT, 0, (LPARAM)&pform2);
	if (!enableLinegap){
		pform2.dyLineSpacing = fontheightTwip*4/3; pform2.bLineSpacingRule = 4;
	} else {
		int		_dyLineSpacing_;
		pProfile->Read("RichEditLinegap", _dyLineSpacing_, 240); pform2.bLineSpacingRule = 4;
		pform2.dyLineSpacing = _dyLineSpacing_;
	}

	// LtwiplɊۂ߂
	HDC		hdcRe = GetDC(Re.Hwnd());
	_ASSERT(hdcRe);
	pxPerInch = GetDeviceCaps(hdcRe, LOGPIXELSY);
	ReleaseDC(Re.Hwnd(), hdcRe);
	int		twipStep = 1440/pxPerInch;
	pform2.dyLineSpacing = (pform2.dyLineSpacing+twipStep-1)/twipStep*twipStep;

	lineheightTwip = pform2.dyLineSpacing;
	pform2.cbSize = sizeof(PARAFORMAT2); pform2.dwMask = PFM_LINESPACING | PFM_SPACEAFTER;
	Re.SendMessage(EM_SETPARAFORMAT, 0, (LPARAM)&pform2);
}

void ItalkView::SetLinegap()
{
    Re.SetSel(0, Re.MoveCaretToTheLast() + 2); SetLinegapSel();
	Re.MoveCaretToTheLast();
}

// WXgǂ
void ItalkView::ReadRegistry(BOOL does_set_current_font)
{
	delete [] color_pattern;

	wchar_t buf[BUFSIZE];
	//lgRegistry registry(REGISTRY_ROOT);
	ProfileManager* pProfile = witalk2parent->getProfileManager();

	// F
	char name[BUFSIZE] = "ColorPattern";
	char *_name = name + strlen(name);
	int i;
	// ܂ ColorPattern%d ̐𐔂
	for (i = 0; ; i++) {
		sprintf(_name, "%d", i);
		if (!pProfile->Read(name, buf, sizeof(buf))) break;
	}
	color_pattern_num = i;
	color_pattern = new ColorPattern[color_pattern_num];
	// ۂ ColorPattern%d ǂݍ
	for (i = 0; i < color_pattern_num; i++) {
		wsprintf(_name, "%d", i);
		pProfile->Read(name, buf, sizeof(buf));
		color_pattern[color_pattern_num - 1 - i].Set(buf);
	}

	// wiF
	Re.SetBackgroundColor(GetColor(*bkColor));

	// tHg
	pProfile->Read("Font", lf, lgWindows::LoadString(rStringDefaultLogFont));
	color_cf.cbSize          = sizeof(color_cf);
	color_cf.dwMask          = 0;
	color_cf.dwEffects       = (((400 < lf.lfWeight) ? CFE_BOLD      : 0) |
		(lf.lfItalic         ? CFE_ITALIC    : 0) |
		(lf.lfStrikeOut      ? CFE_STRIKEOUT : 0) |
		(lf.lfUnderline      ? CFE_UNDERLINE : 0));
	color_cf.yHeight         = abs(lf.lfHeight) * 16;
	color_cf.yOffset         = 0;
	color_cf.crTextColor     = GetColor(*foColor);
	color_cf.bCharSet        = lf.lfCharSet;
	color_cf.bPitchAndFamily = lf.lfPitchAndFamily;
	strcpy(color_cf.szFaceName, lf.lfFaceName);
	if (does_set_current_font) {
		LONG_PTR style = GetWindowLongPtr(Re.Hwnd(), GWL_STYLE);
		SetWindowLongPtr(Re.Hwnd(), GWL_STYLE, style & ~ES_AUTOVSCROLL);

		CHARRANGE backup_selected_cr;
		Re.GetSel(&backup_selected_cr);

		Re.SetSel(0, Re.MoveCaretToTheLast() + 2);
		set_current_font(false);
		Re.MoveCaretToTheLast();
		set_current_font(true);

		SetWindowLongPtr(Re.Hwnd(), GWL_STYLE, (LONG_PTR)style);

	}
	fontheightTwip = color_cf.yHeight;
	pxPerInch = 0;

}

// URL 炵ɒFA
bool ItalkView::detectUrl(const wchar_t* text, const wchar_t* id, int& start_idx, int& end_idx)
{
	const wchar_t*	idx = wcsstr(text, id);
	if (!idx) return false;
	start_idx = (idx - text);
	for (; *idx != 0; idx++) if (*idx >= 0x80 || *idx <= 0x20 ||  *idx == L'<' || *idx == L'>' || *idx == L'"' || *idx == 0x27 || *idx == 0x60 || *idx == L'(' || *idx == L')') break;
	end_idx = idx - text - 1;

	return true;
}

// eLXgԌɑ}
void ItalkView::AddText(const wchar_t *text, BOOL does_autoscroll)
{
	int need_restore = FALSE;
// 0.3.1beta8: ς
//	POINT	org_scrollpos;

	if (!scdata->inputMode && wfocusStatus.isFocused(Re.Hwnd())) does_autoscroll = false;

	int orgCaretPos = Re.GetCaretIndex();

// 0.3.1beta7: ES_AUTOSCROLL ̓IύXsȂ悤ɕύX
//	LONG_PTR style = GetWindowLongPtr(Re.Hwnd(), GWL_STYLE);
	if (does_autoscroll){
// 0.3.1beta7: ES_AUTOSCROLL ̓IύXsȂ悤ɕύX
//	    SetWindowLongPtr(Re.Hwnd(), GWL_STYLE, style | ES_AUTOVSCROLL);
	} else {
// 0.3.1beta7: ES_AUTOSCROLL ̓IύXsȂ悤ɕύX
//		SetWindowLongPtr(Re.Hwnd(), GWL_STYLE, style & ~ES_AUTOVSCROLL);
#if 0
		if (!scdata->inputMode && wfocusStatus.isFocused(Re.Hwnd())){
			SendMessage(Re.Hwnd(), WM_KILLFOCUS, int(witalk2parent->hwnd_combo_connectto), ReSubClass::focusMagicNo);
			need_restore = TRUE;
		} else if (scdata->inputMode && wfocusStatus.isFocused(Re.Hwnd())){
			SendMessage(Re.Hwnd(), WM_KILLFOCUS, int(witalk2parent->hwnd_combo_connectto), ReSubClass::focusMagicNo);
			need_restore = TRUE;
		}
#endif
		if (wfocusStatus.isFocused(Re.Hwnd())) {
// 0.3.1beta8: ς
#if 0
			SendMessage(Re.Hwnd(), EM_GETSCROLLPOS, 0, (LPARAM)&org_scrollpos);
#endif
			SendMessage(Re.Hwnd(), WM_KILLFOCUS, int(witalk2parent->hwnd_combo_connectto), ReSubClass::focusMagicNo);
			need_restore = TRUE;
		}
  }

	CHARRANGE cr_bak;
	Re.GetSel(&cr_bak);

	Re.SuperMoveCaretToTheLast();
	int char_index = Re.GetCaretIndex(); // Why ?

	int len;
	char *mbstext;
	alloca_wcstombs(mbstext, text, len);

	Re.ReplaceSel(text);
	Re.ReleaseSetSel();
	if (does_autoscroll && !need_restore) {	// sɈړ
		Re.SuperMoveCaretToTheLast(); Re.ReleaseSetSel();
	}

	char buf[BUFSIZE];

	//   URL ɉ
	if (DialogURL::url_coloring) {
		int	ul_start = 0, ul_end = 0, searchbase = 0;
		while (true) {
			if      (detectUrl(text+searchbase, L"http://", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else if (detectUrl(text+searchbase, L"ttp://", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else if (detectUrl(text+searchbase, L"https://", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else if (detectUrl(text+searchbase, L"ftp://", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else if (detectUrl(text+searchbase, L"mailto:", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else if (detectUrl(text+searchbase, L"ms-help://", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else if (detectUrl(text+searchbase, L"mms://", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else if (detectUrl(text+searchbase, L"\\\\", ul_start, ul_end)) set_urlunderline(char_index+searchbase+ul_start, char_index+searchbase+ul_end);
			else break;
			searchbase += ul_end;
		}
	}

	// Ft
	for (int i = 0; i < color_pattern_num; i++) {
		const char *str1 = mbstext;
		const char *str2 = buf;
		wcstombs(buf, color_pattern[i].Pattern(), sizeof(buf));
		buf[sizeof(buf) - 1] = '\0';
		wchar_t	cptn[BUFSIZE];
		memset(cptn, 0, sizeof(cptn)); wcscpy(cptn, color_pattern[i].Pattern());

		int char_i = char_index;

		wchar_t	*wc_end = (wchar_t*)text;

		// Waffle 0.2.1:
		//   Windows2000+RichEdit3 ł set_color  Unicode œ삳Ȃ΂ȂȂBDoesMatch  Unicode 
		//   sKvB
		if (*str2 == '^') {
			if (lgString::DoesMatch(text, cptn+1, wc_end)){
				set_color(char_i, char_i + (wc_end - text), color_pattern[i].Color());
			}
		} else {
			while ((unsigned)(wc_end - text) < wcslen(text)){
				if (str1 - mbstext > len) break;
				if (lgString::DoesMatch(text + (str1 - mbstext), cptn, wc_end)){
					set_color(char_i, char_i + (wc_end - (text + (str1 - mbstext))), color_pattern[i].Color());
					char_i += (wc_end - (text + (str1 - mbstext)));
					str1 += (wc_end - (text + (str1 - mbstext)));
				} else {
					wc_end++;
					const char *_str1 = str1;
					str1++;
					char_i += str1 - _str1;
				}
			}
		}
	}

	if (!pxPerInch) {	// s̍擾
		HDC		hdcRe = GetDC(Re.Hwnd());
		_ASSERT(hdcRe);
		pxPerInch = GetDeviceCaps(hdcRe, LOGPIXELSY);
		ReleaseDC(Re.Hwnd(), hdcRe);
	}

// 0.3.1beta7: "&& !need_restore" ǉ
	if (does_autoscroll && !need_restore) {
		RECT	rectre;
		GetClientRect(Re.Hwnd(), &rectre);
		int		caretLines = Re.LineFromChar(Re.GetCaretIndex());
		POINT	movere;
		movere.x = 0;
		if (caretLines < rectre.bottom/(lineheightTwip*pxPerInch/1440)) {
			movere.y = 0;
		} else {
			movere.y = (LONG)(_int64(caretLines - rectre.bottom/(lineheightTwip*pxPerInch/1440))*_int64(lineheightTwip*pxPerInch/1440));
		}
		SendMessage(Re.Hwnd(), EM_SETSCROLLPOS, 0, (LPARAM)&movere);
	} else {
		Re.SetSel(&cr_bak);
		if (need_restore) {
			SendMessage(Re.Hwnd(), WM_SETFOCUS, 0, ReSubClass::focusMagicNo);
// 0.3.1beta8: ς
#if 0
			POINT	cur_scrollpos;
			SendMessage(Re.Hwnd(), EM_GETSCROLLPOS, 0, (LPARAM)&cur_scrollpos);
			if (org_scrollpos.x != cur_scrollpos.x || org_scrollpos.y != cur_scrollpos.y) SendMessage(Re.Hwnd(), EM_SETSCROLLPOS, 0, (LPARAM)&org_scrollpos);
#endif
		}
// 0.3.1beta7: ES_AUTOSCROLL ̓IύXsȂ悤ɕύX
//		SetWindowLong(Re.Hwnd(), GWL_STYLE, style);
	}
#if 0
	// Waffle 0.1.8: ׂ̂pɂǂI
	if (wfocusStatus.isFocused(witalk2parent->iv_log_public->Re.Hwnd())){
		wfocusStatus.killFocus(witalk2parent->iv_input->Re.Hwnd());
	}
#endif
}

void ItalkView::setOldWndProc(WNDPROC wndproc)
{
	old_wnd_proc = wndproc;
}

// Ō̋s̐擪
int ItalkView::MoveCaretToTheLast(void)
{
	int char_index = Re.MoveCaretToTheLast();
	if (char_index != Re.BeginingOfLineCharFromChar(char_index)) {
		Re.Insert(L"\r\n");
		char_index = Re.MoveCaretToTheLast();
	}
	return char_index;
}


// FƂėLȐFԂ
COLORREF ItalkView::GetColor(COLORREF color)
{
	if (color & 0xff000000) return GetSysColor(~color); else return color;
}
