/*

  Quickhack for Italk protocol 1.x that would use ISO-2022
  This code is written by Hidenori Takeshima and
  this code is not covered by GPL;
  you can apply "the ITALK LICENSE" or "LGPL 2.1" for this module.
  If you want to apply LGPL, the version of LGPL MUST be 2.1.

 */

#include <stddef.h>
#include "iso2022.h"


int Quickhack_jis_to_eucjp( char* pdst, char* psrc )
{
	int fJIS = 0;
	int ret = 0;

	while ( *psrc != '\0' )
	{
		if ( *psrc == '\x1b' )
		{
			/* NOTE: ESC ( H should not be used... */
			if ( ( *(psrc+1) == '(' && *(psrc+2) == 'B' ) ||
				 ( *(psrc+1) == '(' && *(psrc+2) == 'J' ) ||
				 ( *(psrc+1) == '(' && *(psrc+2) == 'H' ) )
			{
				fJIS = 0;
				psrc += 3;
				continue;
			}
			else
			if ( ( *(psrc+1) == '$' && *(psrc+2) == '@' ) ||
				 ( *(psrc+1) == '$' && *(psrc+2) == 'B' ) )
			{
				/* map JIS to G0 */
				fJIS = 1;
				psrc += 3;
				continue;
			}

			/* for Kana and JIS X0212 */
			if ( ( *(psrc+1) == '*' && *(psrc+2) == 'I' ) ||
				 ( *(psrc+1) == '+' && *(psrc+2) == 'D' ) )
			{
				psrc += 3;
				continue;
			}
		}

		if ( *psrc >= '\x20' && *psrc <= '\x7e' )
		{
			if ( fJIS && ( *(psrc+1) >= '\x20' && *(psrc+1) <= '\x7e' ) )
			{
				if ( pdst != NULL )
				{
					pdst[ret+0] = psrc[0] | '\x80';
					pdst[ret+1] = psrc[1] | '\x80';
				}
				psrc += 2;
				ret += 2;
			}
			else
			{
				if ( pdst != NULL )
				{
					pdst[ret+0] = psrc[0];
				}
				psrc += 1;
				ret += 1;
			}
		}
		else
		if ( ( *psrc >= '\xa0' && *psrc <= '\xfe' ) &&
			 ( *(psrc+1) >= '\xa0' && *(psrc+1) <= '\xfe' ) )
		{
			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0];
				pdst[ret+1] = psrc[1];
			}
			psrc += 2;
			ret += 2;
		}
		else
		if ( *psrc == '\x8e' )
		{
			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0];
				pdst[ret+1] = psrc[1];
			}
			psrc += 2;
			ret += 2;
		}
		else
		if ( *psrc == '\x8f' )
		{
			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0];
				pdst[ret+1] = psrc[1];
				pdst[ret+2] = psrc[2];
			}
			psrc += 3;
			ret += 3;
		}
		else
		{
			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0];
			}
			psrc += 1;
			ret += 1;
		}

	}

	if ( pdst != NULL )
		pdst[ret+0] = '\0';
	ret += 1;

	return ret;
}

int Quickhack_eucjp_to_jis( char* pdst, char* psrc )
{
	int fJIS = 0;
	int fSetG2 = 0;
	int fSetG3 = 0;
	int ret = 0;

	while ( *psrc != '\0' )
	{
		if ( ( *psrc >= '\xa0' && *psrc <= '\xfe' ) &&
			 ( *(psrc+1) >= '\xa0' && *(psrc+1) <= '\xfe' ) )
		{
			if ( fJIS == 0 )
			{
				if ( pdst != NULL )
				{
					pdst[ret+0] = '\x1b';
					pdst[ret+1] = '$';
					pdst[ret+2] = 'B';
				}
				ret += 3;
				fJIS = 1;
			}

			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0] & (char)0x7f;
				pdst[ret+1] = psrc[1] & (char)0x7f;
			}
			psrc += 2;
			ret += 2;
		}
		else
		if ( *psrc == '\x8e' )
		{
			if ( !fSetG2 )
			{
				if ( pdst != NULL )
				{
					pdst[ret+0] = '\x1b';
					pdst[ret+1] = '*';
					pdst[ret+2] = 'I';
				}
				ret += 3;
				fSetG2 = 1;
			}
			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0];
				pdst[ret+1] = psrc[1];
			}
			psrc += 2;
			ret += 2;
		}
		else
		if ( *psrc == '\x8f' )
		{
			if ( !fSetG3 )
			{
				if ( pdst != NULL )
				{
					pdst[ret+0] = '\x1b';
					pdst[ret+1] = '+';
					pdst[ret+2] = 'D';
				}
				ret += 3;
				fSetG3 = 1;
			}
			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0];
				pdst[ret+1] = psrc[1];
				pdst[ret+2] = psrc[2];
			}
			psrc += 3;
			ret += 3;
		}
		else
		{
			if ( fJIS != 0 )
			{
				if ( pdst != NULL )
				{
					pdst[ret+0] = '\x1b';
					pdst[ret+1] = '(';
					pdst[ret+2] = 'B';
				}
				ret += 3;
				fJIS = 0;
			}

			if ( pdst != NULL )
			{
				pdst[ret+0] = psrc[0];
			}
			psrc += 1;
			ret += 1;
		}
	}

	if ( pdst != NULL )
		pdst[ret+0] = '\0';
	ret += 1;

	return ret;
}
