/*
 * LIBGANA - library for GANA
 * Copyright (C) 1996-1997, TAGA Nayuta <nayuta@is.s.u-tokyo.ac.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

///////////////////////////////////////////////////////////////////////////////
// lgWindows


#include	"lgWindows.h"
#include	"lgRegistry.h"
#include	<locale.h>
#include	<time.h>
#include	<io.h>
#include	<fcntl.h>

#include	<map>
using namespace std;


// libgana ̏
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH) lgWindows::init(hinstDLL);
	else if (fdwReason == DLL_PROCESS_DETACH) lgWindows::end();
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
// lgWindows


HINSTANCE lgWindows::hinst;     // CX^Xnh
HINSTANCE lgWindows::lg_hinst;  // libgana  hinst
BOOL      lgWindows::is_nt_351; // OS  NT3.51 ǂ
BOOL      lgWindows::is_nt;     // OS  NT ǂ


// 
BOOL lgWindows::init(HINSTANCE _lg_hinst)
{
  lgWindows::hinst    = GetModuleHandle(NULL);
  lgWindows::lg_hinst = _lg_hinst;
  
  // RRg[̏
  InitCommonControls();
  
  // Windows o[W
  OSVERSIONINFO ver;
  ver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx(&ver);
  lgWindows::is_nt     = ver.dwPlatformId == VER_PLATFORM_WIN32_NT;
  lgWindows::is_nt_351 = ver.dwMajorVersion == 3 && is_nt;
  
  // local ܂̐ݒ
  setlocale(LC_ALL, "japanese"); // P[̐ݒ
  if (!getenv("TZ"))
  {
    // VC++  C CũoO ?
    _putenv("TZ=JST-9");
    _tzset();
  }
  
  // lgVersion ̏
  lgVersion::init();
  
  // lgCtl3d ̏
  lgCtl3d::init();
  
  return TRUE;
}


// Ƃ܂
BOOL lgWindows::end(void)
{
	// lgVersion ̂Ƃ܂
	lgVersion::end();

	// lgCtl3d ̂Ƃ܂
	lgCtl3d::end();

	return TRUE;
}


// O
const char *lgWindows::TheName(void)
{
	return "WItalk2";
//	return lgVersion::GetVersionString(lgVersion::ProductName);
}


map<int,char*>*		lgWindows::resStringA = NULL;
map<int,wchar_t*>*	lgWindows::resStringW = NULL;
HANDLE				lgWindows::resStringMutexA = NULL;
HANDLE				lgWindows::resStringMutexW = NULL;

// LoadString ̎g₷`
char* lgWindows::LoadString(int id)
{
#if 0
	__declspec(thread) static char buf[BUFSIZE];
	if (::LoadStringA(lgWindows::hInst(), id, buf, sizeof(buf))) return buf; else return NULL;
#else
	if (!resStringMutexA) resStringMutexA = CreateMutex(NULL, FALSE, NULL);
	WaitForSingleObject(resStringMutexA, INFINITE);
	if (!resStringA) resStringA = new map<int,char*>;
	map<int,char*>::iterator p = resStringA->find(id);
	if (p == resStringA->end()) {
		char	buf[BUFSIZE];
		if (!::LoadStringA(lgWindows::hInst(), id, buf, sizeof(buf))) { ReleaseMutex(resStringMutexA); return NULL; }
		char*	resstring = new char[strlen(buf)+1];
		strcpy(resstring, buf);
		resStringA->insert(pair<int,char*>(id, resstring));
		p = resStringA->find(id);
	}
	ReleaseMutex(resStringMutexA);
	return p->second;
#endif
}

wchar_t* lgWindows::LoadStringW(int id)
{
#if 0
	__declspec(thread) static wchar_t buf[BUFSIZE];
	if (::LoadStringW(lgWindows::hInst(), id, buf, sizeof(buf))) return buf; else return NULL;
#else
	if (!resStringMutexW) resStringMutexW = CreateMutex(NULL, FALSE, NULL);
	WaitForSingleObject(resStringMutexW, INFINITE);
	if (!resStringW) resStringW = new map<int,wchar_t*>;
	map<int,wchar_t*>::iterator p = resStringW->find(id);
	if (p == resStringW->end()) {
		wchar_t	buf[BUFSIZE];
		if (!::LoadStringW(lgWindows::hInst(), id, buf, sizeof(buf))) {
			char*	resA = LoadString(id);
			if (!resA) {ReleaseMutex(resStringMutexW); return NULL; }
			mbstowcs(buf, resA, strlen(resA)+1);
		}
		wchar_t*	resstring = new wchar_t[wcslen(buf)+1];
		wcscpy(resstring, buf);
		resStringW->insert(pair<int,wchar_t*>(id, resstring));
		p = resStringW->find(id);
	}
	ReleaseMutex(resStringMutexW);
	return p->second;
#endif
}


// ChooseColor ̊ȈՔ
BOOL lgWindows::ChooseColor(HWND hwnd_owner, COLORREF &color)
{
	CHOOSECOLOR cc;
	ZeroMemory(&cc, sizeof(cc));
	DWORD custcolor[16], size_cc = sizeof(custcolor);
	ZeroMemory(custcolor, sizeof(custcolor));
	lgRegistry registry(HKEY_CURRENT_USER, "Control Panel\\Appearance");
	registry.Read("CustomColors", (BYTE *)custcolor, size_cc);
	cc.lStructSize = sizeof(cc);
	cc.hwndOwner = hwnd_owner;
	cc.rgbResult = color;
	cc.lpCustColors = custcolor;
	cc.Flags = CC_FULLOPEN | CC_RGBINIT;
	if (::ChooseColor(&cc)) {
		registry.Write("CustomColors", (BYTE *)custcolor, size_cc);
		color = cc.rgbResult;
		return TRUE;
	} else {
		return FALSE;
	}
}


// WR\[쐬
BOOL lgWindows::AllocStdConsole(BOOL text_mode)
{

	// Waffle 0.1.3: Warning 
#if 0
	if (!AllocConsole())
    /* return FALSE */;
#else
	AllocConsole();
#endif

  int flags = text_mode ? _O_TEXT : 0;
  
  FILE *fp;
  fp = _fdopen(_open_osfhandle(
    (long)GetStdHandle(STD_INPUT_HANDLE), flags), "r");
  if (fp)
    *stdin = *fp;
  fp = _fdopen(_open_osfhandle(
    (long)GetStdHandle(STD_OUTPUT_HANDLE), flags), "w");
  if (fp)
    *stdout = *fp;
  fp = _fdopen(_open_osfhandle(
    (long)GetStdHandle(STD_ERROR_HANDLE), flags), "w");
  if (fp)
    *stderr = *fp;
  
  return TRUE;
}


// ImageList ǂݍ
HIMAGELIST LoadImageList(int rBitmap16, int rBitmap256, int width)
{
  if (16 <= lgGDI::BitsParPixel() && 0 <= rBitmap256 && !lgWindows::IsNT())
    return ImageList_LoadImage(lgWindows::hInst(),
			       MAKEINTRESOURCE(rBitmap256),
			       width, 0, CLR_NONE, IMAGE_BITMAP,
			       LR_LOADTRANSPARENT | LR_CREATEDIBSECTION);
  else
    return ImageList_LoadImage(lgWindows::hInst(),
			       MAKEINTRESOURCE(rBitmap16),
			       width, 0, CLR_NONE, IMAGE_BITMAP,
			       LR_LOADTRANSPARENT | LR_CREATEDIBSECTION);
}


///////////////////////////////////////////////////////////////////////////////
// lgVersion


char *lgVersion::lg_version_buf; // libgana.dll ̃o[W
char *lgVersion::version_buf;    // dll Ăяoo[W
const char *lgVersion::vc[9] =             // o[WJeS
{
  "\\StringFileInfo\\041104b0\\Comments",
  "\\StringFileInfo\\041104b0\\CompanyName",
  "\\StringFileInfo\\041104b0\\FileDescription",
  "\\StringFileInfo\\041104b0\\FileVersion",
  "\\StringFileInfo\\041104b0\\InternalName",
  "\\StringFileInfo\\041104b0\\LegalCopyright",
  "\\StringFileInfo\\041104b0\\OriginalFilename",
  "\\StringFileInfo\\041104b0\\ProductName",
  "\\StringFileInfo\\041104b0\\ProductVersion",
};


// o[W̓ǂݍ
BOOL lgVersion::init(void)
{
  char buf[BUFSIZE];
  int sizeof_version_buf;
  
  // dll Ăяoo[W
  GetModuleFileName(lgWindows::hInst(), buf, sizeof(buf));
  sizeof_version_buf = GetFileVersionInfoSize(buf, NULL);
  if (sizeof_version_buf)
  {
    version_buf = new char[sizeof_version_buf];
    if (!GetFileVersionInfo(buf, NULL, sizeof_version_buf, version_buf))
    {
      delete [] version_buf;
      version_buf = NULL;
    }
  }
  
  // libgana.dll ̃o[W  
  GetModuleFileName(lgWindows::lg_hInst(), buf, sizeof(buf));
  sizeof_version_buf = GetFileVersionInfoSize(buf, NULL);
  if (sizeof_version_buf)
  {
    lg_version_buf = new char[sizeof_version_buf];
    if (!GetFileVersionInfo(buf, NULL, sizeof_version_buf, lg_version_buf))
    {
      delete [] lg_version_buf;
      lg_version_buf = NULL;
    }
  }
  return TRUE;
}


// lgVersion ̂Ƃ܂
BOOL lgVersion::end(void)
{
  if (version_buf)
    delete [] version_buf;
  if (lg_version_buf)
    delete [] lg_version_buf;
  return TRUE;
}


// o[W
char *lgVersion::GetVersionString(lgVersion::VersionCategory v,
				  const char *_version_buf)
{
  if (!_version_buf)
    return NULL;
  char *data;
  UINT data_size;
  VerQueryValue((void *)_version_buf, (char *)(vc[v]),
		(LPVOID *)&data, &data_size);
  return data;
}


// ̃\tgEFÃt@Co[W
BOOL lgVersion::GetVersionNo(int *v, const char *_version_buf)
{
  char *vs = GetVersionString(FileVersion, _version_buf);
  if (!vs)
    return FALSE;
  for (int i = 0; i < 4; i++)
  {
    v[i] = atoi(vs);
    vs = _mbschr(vs, ',');
    if (!vs)
      return FALSE;
    vs += 2;
  }
  return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
// lgCtl3d


HMODULE lgCtl3d::hmodule;
lgCtl3d::_register      lgCtl3d::Register;
lgCtl3d::_auto_subclass lgCtl3d::AutoSubclass;
lgCtl3d::_color_change  lgCtl3d::ColorChange;
lgCtl3d::_unregister    lgCtl3d::Unregister;


// NT3.51pCTL3D32 (pς)
BOOL lgCtl3d::init(void)
{
	return TRUE;
}


// Ƃ܂
BOOL lgCtl3d::end(void)
{
  if (hmodule)
  {
    if (Unregister)
      Unregister(lgWindows::hInst());
    FreeLibrary(hmodule);
  }
  hmodule = NULL;
  return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
// lgGDI


// pbgĂ邩ǂ
BOOL lgGDI::HasPalette(void)
{
  HDC hdc = GetDC(GetDesktopWindow());
  BOOL r = GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE;
  ReleaseDC(GetDesktopWindow(), hdc);
  return r;
}


// sNZ̃rbg
int lgGDI::BitsParPixel(void)
{
  HDC hdc = GetDC(GetDesktopWindow());
  int bits_par_pixel = GetDeviceCaps(hdc, BITSPIXEL);
  ReleaseDC(GetDesktopWindow(), hdc);
  return bits_par_pixel;
}
