/*
 * LIBGANA - library for GANA
 * Copyright (C) 1996-1997, TAGA Nayuta <nayuta@is.s.u-tokyo.ac.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

///////////////////////////////////////////////////////////////////////////////
// lgcRichEditEmacsBinding.cxx


#include	<crtdbg.h>
#include	"Trace.h"
#include	"lgcRichEdit.h"
#include	"lgList.h"
#include	"WinDLL.h"

const int WM_APP_RUN = (WM_APP + 200);


static LRESULT CALLBACK EmacsBindingWndProc(HWND hwnd, UINT message,
					    WPARAM wParam, LPARAM lParam);


class EmacsBinding : lgcRichEdit
{
	KeyBinding key_binding;           // Emacs  Windows ŋL[Xg
	//                                // [N̏ꍇAǂD悷邩
	lgcStatusbar *statusbar;          // Isearch ̓e\Xe[^Xo[
	int isearch_mode;                 // -1:backward, 0:non-isearch, 1:forward
	char isearch_word[BUFSIZE];       // ̒P
	char pre_isearch_word[BUFSIZE];   // ǑP
	int last_index;                   // RichEdit ̈ԍŌ̈ʒu
	BOOL is_precommand_kill;          // ȑOɎsR}h kill nǂ
	
	lgKeyBindingsMap emacs_keymap;
	lgKeyBindingsMap windows_keymap;
	lgKeyBindingsMap ctrl_c_keymap;
	lgKeyBindingsMap ctrl_x_keymap;
	lgKeyBindingsMap isearch_keymap;
	lgKeyBindingsMap *current_keymap;


	enum {
		Shift   = lgKeyBindingsMap::Shift,
		Control = lgKeyBindingsMap::Control,
		Ctrl    = lgKeyBindingsMap::Ctrl,
		Alt     = lgKeyBindingsMap::Alt,
	};

	// Isearch ɌꏊۑNX
	class IsearchPlace
	{
	public:
		const int Index;  // ꏊ
		const int Length; // P̒
		const int Mode;   // 0:forward, 1:backward, 2:forward fail, 3:backward fail
		IsearchPlace(int index, int length, int mode = 0)
			: Index(index), Length(length), Mode(mode) { }
		IsearchPlace(const IsearchPlace &isp)
			: Index(isp.Index), Length(isp.Length), Mode(isp.Mode) { }
	};
	lgList<IsearchPlace> isearch_place; // ꏊ
	
	// Isearch Ɍꏊ̂AԍŌɌ̂Ԃ
	IsearchPlace &last_place(void)
	{
		return isearch_place[isearch_place.Length() - 1];
	}

private:
	// Isearch ̃Xe[^X\
	void isearch_show_status(int is_failed = FALSE)
	{
		if (statusbar)
			if (isearch_mode == 0) statusbar->SetDefaultText("");
			else {
				char buf[sizeof(isearch_word) + 100];
				IsearchPlace &lp = last_place();
				if ((lp.Mode & 2) || is_failed)
					strcpy(buf, "Failing ");
				else {
					AQT(("lp.Index=%d, lp.Length=%d\n", lp.Index, lp.Length));
					SetSel(lp.Index, lp.Index + lp.Length);
					buf[0] = '\0';
				}
				strcat(buf, (lp.Mode & 1) ? "I-search backward: " : "I-search: ");
				strcat(buf, isearch_word);
				statusbar->SetDefaultText(buf);
			}
	}
  
	// Isearch n߂
	void isearch_start(int i)
	{
		if (isearch_mode != 0) return;
		isearch_mode = i;
    
		// RichEdit ̈ԍŌ̈ʒuۑ
		last_index = CharFromLine(GetLineCount() - 1);
		last_index += GetLineLengthFromChar(last_index);
    
		// O񌟍ꏊ폜
		while (isearch_place.Length()) isearch_place.Delete(0);
		
		// n߂ꏊۑ
		isearch_place.Add(IsearchPlace(GetCaretIndex(), 0, (isearch_mode == -1)));
		isearch_word[0] = '\0';
		current_keymap = &isearch_keymap;

		isearch_show_status();
	}

	// ǉ
	void isearch_string(char *str)
	{
		strcat(isearch_word, str);
		isearch_show_status(!isearch_search());
	}

	// T[`
	BOOL isearch_search(BOOL does_add_isearch_place_when_failed = TRUE)
	{
		if (isearch_mode == 0) return FALSE;
		
		// ʒuƏꏊ
		FINDTEXT ft;
		GetSel(&(ft.chrg));
		ft.lpstrText = isearch_word;
		
		if (does_add_isearch_place_when_failed == FALSE)
			if (isearch_mode == 1) 
				ft.chrg.cpMin = ft.chrg.cpMax;
			else
				ft.chrg.cpMax = ft.chrg.cpMin;
			
		int i = -1;
		if (isearch_mode == 1) { // O
			ft.chrg.cpMax = last_index;
			AQT(("pre: %d %d %d %d %s\n", i, ft.chrg.cpMin, ft.chrg.cpMax, last_index, ft.lpstrText));
			i = FindTextA(&ft, FR_DOWN | FR_MATCHCASE);
			AQT(("aft: %d %d %d %d %s\n", i, ft.chrg.cpMin, ft.chrg.cpMax, last_index, ft.lpstrText));
		} else { // 
			ft.chrg.cpMax = last_index;
			AQT(("pre: %d %d %d %d %s\n", i, ft.chrg.cpMin, ft.chrg.cpMax, last_index, ft.lpstrText));
			i = FindTextA(&ft, FR_MATCHCASE);
			AQT(("aft: %d %d %d %d %s\n", i, ft.chrg.cpMin, ft.chrg.cpMax, last_index, ft.lpstrText));
		}

		int		wordlen_ucs2 = mbstowcs(NULL, ft.lpstrText, strlen(ft.lpstrText));
		_ASSERT(wordlen_ucs2 != -1);

		if (0 <= i) { // 
			isearch_place.Add(IsearchPlace(i, wordlen_ucs2, (isearch_mode == -1)));
			return TRUE;
		} else { // s
			if (does_add_isearch_place_when_failed)
				isearch_place.Add(IsearchPlace(last_place().Index, wordlen_ucs2, 2 | (int)(isearch_mode == -1)));
			return FALSE;
		}
	}

	// Nbv{[hɕݒ肷
	void set_string_to_clipboard(const char *str) {
		if (OpenClipboard(NULL)) {
			EmptyClipboard();
			HGLOBAL hdata = GlobalAlloc(GMEM_MOVEABLE, strlen(str) + 1);
			char *data = (char *)GlobalLock(hdata);
			_ASSERT(data);
			strcpy(data, str);
			GlobalUnlock(hdata);
			SetClipboardData(CF_TEXT, hdata);
			CloseClipboard();
		}
	}

	// Nbv{[h當擾
	char *new_string_from_clipboard(void)
	{
		char *r = NULL;
		if (OpenClipboard(NULL)) {
			HGLOBAL hdata = GetClipboardData(CF_TEXT);
			if (hdata) {
				char *data = (char *)GlobalLock(hdata);
				r = new_string(data);
				GlobalUnlock(hdata);
			}
			CloseClipboard();
		}
		return r;
	}
	
	// IWi wndproc ɃL[𑗂
	void send_vkey(int vkey, int shift = 0)
	{
		BYTE bak[256], state[256];
		GetKeyboardState(bak);
		GetKeyboardState(state);
		state[VK_RCONTROL] = state[VK_LCONTROL] = state[VK_CONTROL] =
			(shift & Control) ? 0x80 : 0;
		state[VK_RSHIFT] = state[VK_LSHIFT] = state[VK_SHIFT] =
			(shift & Shift) ? 0x80 : 0;
		state[VK_RMENU] = state[VK_LMENU] = state[VK_MENU] =
			(shift & Alt) ? 0x80 : 0;
		SetKeyboardState(state);
		CallWindowProc((WNDPROC)original_wndproc, Hwnd(), WM_KEYDOWN, vkey,
			((255 & MapVirtualKey(vkey, 0)) << 16));
		CallWindowProc((WNDPROC)original_wndproc, Hwnd(), WM_KEYUP, vkey,
			((255 & MapVirtualKey(vkey, 0)) << 16) | 0x80000000);
		SetKeyboardState(bak);
	}

public:
	
	// grep 'run_[a-zA-Z][a-zA-Z]*' lgcRichEditEmacsBinding.cxx | grep -v 'run_[Cc]ommand' | sed 's/^.*\(run_[a-zA-Z][a-zA-Z]*\).*$/    \1,/' | sort | uniq
	
	enum run_Commands {
		run_BackwardChar,
		run_BackwardKillWord,
		run_BackwardWord,
		run_BeginningOfBuffer,
		run_BeginningOfLine,
		run_ControlCMap,
		run_ControlXMap,
		run_DeleteBackwardChar,
		run_DeleteChar,
		run_DeleteRegionOrKillLine,
		run_EndOfBuffer,
		run_EndOfLine,
		run_ExpandRegionBackwardChar,
		run_ExpandRegionBeginningOfLine,
		run_ExpandRegionEndOfLine,
		run_ExpandRegionForwardChar,
		run_ExpandRegionNextLine,
		run_ExpandRegionPreviousLine,
		run_ExpandRegionScrollDown,
		run_ExpandRegionScrollUp,
		run_ForwardChar,
		run_ForwardWord,
		run_ImeToggle,
		run_IsearchAbort,
		run_IsearchBackward,
		run_IsearchDeleteChar,
		run_IsearchExit,
		run_IsearchExitAndRunChar,
		run_IsearchExitAndRunVKey,
		run_IsearchForward,
		run_IsearchPrintingChar,
		run_IsearchRepeatBackward,
		run_IsearchRepeatForward,
		run_IsearchYank,
		run_KeyboardQuit,
		run_KillLine,
		run_KillRegion,
		run_KillRingSave,
		run_KillWord,
		run_Newline,
		run_NextLine,
		run_OpenLine,
		run_PreviousLine,
		run_ScrollDown,
		run_ScrollUp,
		run_SelectAll,
		run_SelfInsertCommand,
		run_SetMarkCommand,
		run_Undo,
		run_Yank,
	};
	
	int wmAppRun(int run_command, LPARAM lParam)
	{
#define Run(c) case run_##c: c(lParam); return 0;
		// grep 'run_[a-zA-Z][a-zA-Z]*' lgcRichEditEmacsBinding.cxx | grep -v 'run_[Cc]ommand' | sed 's/^.*run_\([a-zA-Z][a-zA-Z]*\).*$/      Run(\1);/' | sort | uniq
		switch (run_command) {
			Run(BackwardChar);
			Run(BackwardKillWord);
			Run(BackwardWord);
			Run(BeginningOfBuffer);
			Run(BeginningOfLine);
			Run(ControlCMap);
			Run(ControlXMap);
			Run(DeleteBackwardChar);
			Run(DeleteChar);
			Run(DeleteRegionOrKillLine);
			Run(EndOfBuffer);
			Run(EndOfLine);
			Run(ExpandRegionBackwardChar);
			Run(ExpandRegionBeginningOfLine);
			Run(ExpandRegionEndOfLine);
			Run(ExpandRegionForwardChar);
			Run(ExpandRegionNextLine);
			Run(ExpandRegionPreviousLine);
			Run(ExpandRegionScrollDown);
			Run(ExpandRegionScrollUp);
			Run(ForwardChar);
			Run(ForwardWord);
			Run(ImeToggle);
			Run(IsearchAbort);
			Run(IsearchBackward);
			Run(IsearchDeleteChar);
			Run(IsearchExit);
			Run(IsearchExitAndRunChar);
			Run(IsearchExitAndRunVKey);
			Run(IsearchForward);
			Run(IsearchPrintingChar);
			Run(IsearchRepeatBackward);
			Run(IsearchRepeatForward);
			Run(IsearchYank);
			Run(KeyboardQuit);
			Run(KillLine);
			Run(KillRegion);
			Run(KillRingSave);
			Run(KillWord);
			Run(Newline);
			Run(NextLine);
			Run(OpenLine);
			Run(PreviousLine);
			Run(ScrollDown);
			Run(ScrollUp);
			Run(SelectAll);
			Run(SelfInsertCommand);
			Run(SetMarkCommand);
			Run(Undo);
			Run(Yank);
		}
		return 0;
#undef Run 
	}

	void ControlCMap(LPARAM code)
	{
		current_keymap->SetTmpMap(&ctrl_c_keymap);
	}

	void ControlXMap(LPARAM code)
	{
		current_keymap->SetTmpMap(&ctrl_x_keymap);
	}

	void BackwardChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_LEFT);
	}

	void BackwardKillWord(LPARAM code)
	{
		send_vkey(VK_LEFT, Control | Shift);
		KillRegion(-1);
	}
  
	void BackwardWord(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_LEFT, Control);
	}

	void BeginningOfBuffer(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_HOME, Control);
		//MoveCaret(0);
	}

	void BeginningOfLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		MoveCaret(BeginingOfLineCharFromChar(GetCaretIndex()));
	}

	void DeleteBackwardChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_BACK);
		/*    int s, e;
		GetSel(&s, &e);
		if (s == e)
		SetSel(s - 1, e);
		GetSel(&s, &e);
		char *str = (char *)alloca(e - s + 2);
		GetSelText(str);
		statusbar->SetText(str);
		Clear();*/
	}
  
	void DeleteChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_DELETE);
	}

	void DeleteRegionOrKillLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_DELETE, Control);
	}

	void EndOfBuffer(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_END, Control);
		//MoveCaretToTheLast();
	}

	void EndOfLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		MoveCaret(EndOfLineCharFromChar(BeginingOfLineCharFromChar(GetCaretIndex())));
	}

	void ExpandRegionBackwardChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_LEFT, Shift);
	}

	void ExpandRegionBeginningOfLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_HOME, Shift);
	}

	void ExpandRegionEndOfLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_END, Shift);
	}
  
	void ExpandRegionForwardChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_RIGHT, Shift);
	}

	void ExpandRegionNextLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_DOWN, Shift);
	}

	void ExpandRegionPreviousLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_UP, Shift);
	}
  
	void ExpandRegionScrollDown(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_PRIOR, Shift);
	}

	void ExpandRegionScrollUp(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_NEXT, Shift);
	}

	void ForwardChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_RIGHT);
	}

	void ForwardWord(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_RIGHT, Control);
	}
  
	void ImeToggle(LPARAM code)
	{
		is_precommand_kill = FALSE;
		HIMC hIMC = ImmGetContext(Hwnd());
		if (hIMC) {
			ImmSetOpenStatus(hIMC, !ImmGetOpenStatus(hIMC));
			ImmReleaseContext(Hwnd(), hIMC);
		}
	}

	void IsearchAbort(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode == 0) return;

		MoveCaret(isearch_place[0].Index);
		isearch_mode = 0;
		current_keymap = (key_binding == KeyBindingEmacs) ? &emacs_keymap : &windows_keymap;
		isearch_show_status();
	}

	void IsearchBackward(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode != 0) IsearchRepeatBackward(code); else isearch_start(-1);
	}

	void IsearchDeleteChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode == 0) return;

		if (1 < isearch_place.Length()) {
			int len = isearch_place.Length();
			isearch_place.Delete(len - 1);
			isearch_word[isearch_place[len - 2].Length] = '\0';
			isearch_mode = (isearch_place[len - 2].Mode & 1) ? -1 : 1;
		}
		isearch_show_status();
	}

	void IsearchExit(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode == 0) return;

		strcpy(pre_isearch_word, isearch_word);
		MoveCaret(last_place().Index + ((isearch_mode == 1) ? last_place().Length : 0));
		isearch_mode = 0;
		current_keymap = (key_binding == KeyBindingEmacs) ? &emacs_keymap : &windows_keymap;
		isearch_show_status();
	}
  
	void IsearchExitAndRunChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		IsearchExit(code);
		current_keymap->Char((char)code);
	}

	void IsearchExitAndRunVKey(LPARAM code)
	{
		is_precommand_kill = FALSE;
		IsearchExit(code);
		current_keymap->KeyDown((int)code);
	}
  
	void IsearchForward(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode != 0) IsearchRepeatForward(code); else isearch_start(1);
	}

	void IsearchPrintingChar(LPARAM code)
	{
		is_precommand_kill = FALSE;
		char buf[] = { (char)code, '\0' };
		isearch_string(buf);
	}

	void IsearchRepeatBackward(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode == 0) return;
    
		isearch_mode = -1;
		BOOL d = TRUE;
		if (isearch_place.Length() == 1 && pre_isearch_word[0] != '\0') strcpy(isearch_word, pre_isearch_word);
		else d = FALSE;
		isearch_show_status(!isearch_search(d));
	}

	void IsearchRepeatForward(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode == 0) return;

		isearch_mode = 1;
		BOOL d = TRUE;
		if (isearch_place.Length() == 1 && pre_isearch_word[0] != '\0') strcpy(isearch_word, pre_isearch_word);
		else d = FALSE;
		isearch_show_status(!isearch_search(d));
	}

	void IsearchYank(LPARAM code)
	{
		is_precommand_kill = FALSE;
		if (isearch_mode == 0) return;
		char *str = new_string_from_clipboard();
		isearch_string(str);
		delete_string(str);
	}
  
	void KeyboardQuit(LPARAM code)
	{
		is_precommand_kill = FALSE;
		IsearchExit(code);
		current_keymap = (key_binding == KeyBindingEmacs) ? &emacs_keymap : &windows_keymap;
		isearch_show_status();
	}

	void KillLine(LPARAM code)
	{
		int end_of_line = EndOfLineCharFromChar(
			BeginingOfLineCharFromChar(GetCaretIndex()));

		int s, e;
		if (end_of_line == GetCaretIndex()) s = end_of_line, e = end_of_line + 2;
		else s = GetCaretIndex(), e = end_of_line;

		SetSel(s, e);

		KillRegion(code);
	}
  
	void KillRegion(LPARAM code)
	{
		KillRingSave(code);
		Delete();
	}

	void KillRingSave(LPARAM code)
	{
		char *pre_str;
		// Waffle 0.2.2beta7: ^C Ałŕ񂪘Aŉ񐔕AăNbv{[ĥ֑C (by Kit)
		//
		if (is_precommand_kill && key_binding == KeyBindingEmacs) pre_str = new_string_from_clipboard();
		else pre_str = new_string("");

		int s, e;
		GetSel(&s, &e);
		size_t prestr_len = strlen(pre_str);
		int		len;
		len = (e - s + 1)*2;		// }u
		char *str = new char[prestr_len + len + 1];
		if (0 <= code) {
			strcpy(str, pre_str);
			GetSelText(str + prestr_len);
		} else {
			GetSelText(str);
			strcat(str, pre_str);
		}
		// RichEdit2EM_GETSELTEXT͍sCR̕Ԃ̂ōsϊKv
		// WM_COPYłRichEditŕϊsăNbv{[hɓ]
		char*	cbtext = NULL;
		{
			int		lns = 0;
			char	*p, *q;
			for (p = str; *p; p++) if (*p == 0x0D && *(p+1) != 0x0A) lns++;
			cbtext = new char[p - str + lns + 1];
			for (p = str, q = cbtext; *p; p++) {
				*(q++) = *p;
				if (*p == 0x0D && *(p+1) != 0x0A) *(q++) = 0x0A;
			}
			*q = 0;
		}

		set_string_to_clipboard(cbtext);

		delete [] cbtext;
		delete_string(str);
		delete_string(pre_str);
		is_precommand_kill = TRUE;
	}

	void KillWord(LPARAM code)
	{
		send_vkey(VK_RIGHT, Control | Shift);
		KillRegion(code);
	}

	void Newline(LPARAM code)
	{
		is_precommand_kill = FALSE;
		CallWindowProc((WNDPROC)original_wndproc, Hwnd(), WM_CHAR, 'M' - '@', 0);
	}
  
	void NextLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_DOWN);
	}

	void OpenLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		Newline(code); BackwardChar(code);
	}

	void PreviousLine(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_UP);
	}
  
	void ScrollDown(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_PRIOR);
	}

	void ScrollUp(LPARAM code)
	{
		is_precommand_kill = FALSE;
		send_vkey(VK_NEXT);
	}

	void SelectAll(LPARAM code)
	{
		is_precommand_kill = FALSE;
		CallWindowProc((WNDPROC)original_wndproc, Hwnd(), WM_CHAR, 'A' - '@', 0);
	}

	void SelfInsertCommand(LPARAM code)
	{
		is_precommand_kill = FALSE;
		CallWindowProc((WNDPROC)original_wndproc, Hwnd(), WM_CHAR, code, 0);
	}
  
	void SetMarkCommand(LPARAM code)
	{
		is_precommand_kill = FALSE;
	}

	void Undo(LPARAM code)
	{
		is_precommand_kill = FALSE;
		lgcRichEdit::Undo();
	}

	void Yank(LPARAM code)
	{
		is_precommand_kill = FALSE;
		Paste();
	}
    
public:
	static WNDPROC original_wndproc; // RichEdit ̃IWi WNDPROC

	EmacsBinding(HWND hwnd, lgcStatusbar *_statusbar)
		: lgcRichEdit(hwnd),
          key_binding(KeyBindingEmacs),
		  statusbar(_statusbar),
		  isearch_mode(FALSE),
		  is_precommand_kill(FALSE),
		  emacs_keymap(hwnd),
		  windows_keymap(hwnd),
		  ctrl_c_keymap(hwnd),
		  ctrl_x_keymap(hwnd),
		  isearch_keymap(hwnd),
		  current_keymap(&emacs_keymap)
	{
		isearch_word[0] = pre_isearch_word[0] = '\0';

		const int self  = lgKeyBindingsMap::DestSelf;
		const int code  = self | lgKeyBindingsMap::lParamCode;

		// emacs
		emacs_keymap.SetCharOver(0x20, 0x7e, 0, code,  WM_APP_RUN, run_SelfInsertCommand);
		emacs_keymap.SetCharOver(0xa1, 0xdf, 0, code,  WM_APP_RUN, run_SelfInsertCommand);
#define esc emacs_keymap.SetChar
		esc('@', Ctrl, self, WM_APP_RUN, run_SetMarkCommand);
		esc('A', Ctrl, self, WM_APP_RUN, run_BeginningOfLine);
		esc('B', Ctrl, self, WM_APP_RUN, run_BackwardChar);
		esc('C', Ctrl, self, WM_APP_RUN, run_ControlCMap);
		esc('D', Ctrl, self, WM_APP_RUN, run_DeleteChar);
		esc('E', Ctrl, self, WM_APP_RUN, run_EndOfLine);
		esc('F', Ctrl, self, WM_APP_RUN, run_ForwardChar);
		esc('G', Ctrl, self, WM_APP_RUN, run_KeyboardQuit);
		//esc('H', Ctrl, self, WM_APP_RUN, run_DeleteBackwardChar); // oOH
		esc('I', Ctrl, code, WM_APP_RUN, run_SelfInsertCommand);
		esc('J', Ctrl, self, WM_APP_RUN, run_Newline);
		esc('K', Ctrl, self, WM_APP_RUN, run_KillLine);
		esc('M', Ctrl, self, WM_APP_RUN, run_Newline);
		esc('N', Ctrl, self, WM_APP_RUN, run_NextLine);
		esc('O', Ctrl, self, WM_APP_RUN, run_OpenLine);
		esc('P', Ctrl, self, WM_APP_RUN, run_PreviousLine);
		esc('R', Ctrl, self, WM_APP_RUN, run_IsearchBackward);
		esc('S', Ctrl, self, WM_APP_RUN, run_IsearchForward);
		esc('V', Ctrl, self, WM_APP_RUN, run_ScrollUp);
		esc('W', Ctrl, self, WM_APP_RUN, run_KillRegion);
		esc('X', Ctrl, self, WM_APP_RUN, run_ControlXMap);
		esc('Y', Ctrl, self, WM_APP_RUN, run_Yank);
		esc('\\',Ctrl, self, WM_APP_RUN, run_ImeToggle);
		esc('_', Ctrl, self, WM_APP_RUN, run_Undo);
		esc('F', Alt , self, WM_APP_RUN, run_ForwardWord);
		esc('f', Alt , self, WM_APP_RUN, run_ForwardWord);
		esc('B', Alt , self, WM_APP_RUN, run_BackwardWord);
		esc('b', Alt , self, WM_APP_RUN, run_BackwardWord);
		esc('D', Alt , self, WM_APP_RUN, run_KillWord);
		esc('d', Alt , self, WM_APP_RUN, run_KillWord);
		esc('V', Alt , self, WM_APP_RUN, run_ScrollDown);
		esc('v', Alt , self, WM_APP_RUN, run_ScrollDown);
		esc('W', Alt , self, WM_APP_RUN, run_KillRingSave);
		esc('w', Alt , self, WM_APP_RUN, run_KillRingSave);
#undef esc
#define esv emacs_keymap.SetVKey
		esv(VK_UP,     0,     self, WM_APP_RUN, run_PreviousLine);
		esv(VK_DOWN,   0,     self, WM_APP_RUN, run_NextLine);
		esv(VK_LEFT,   0,     self, WM_APP_RUN, run_BackwardChar);
		esv(VK_RIGHT,  0,     self, WM_APP_RUN, run_ForwardChar);
		esv(VK_HOME,   0,     self, WM_APP_RUN, run_BeginningOfBuffer);
		esv(VK_END,    0,     self, WM_APP_RUN, run_EndOfBuffer);
		esv(VK_DELETE, 0,     self, WM_APP_RUN, run_DeleteChar);
		esv(VK_NEXT,   0,     self, WM_APP_RUN, run_ScrollUp);
		esv(VK_PRIOR,  0,     self, WM_APP_RUN, run_ScrollDown);
		esv(VK_BACK,   Alt,   self, WM_APP_RUN, run_BackwardKillWord);
		esv(VK_LEFT,   Alt,   self, WM_APP_RUN, run_BackwardWord);
		esv(VK_RIGHT,  Alt,   self, WM_APP_RUN, run_ForwardWord);
		esv(VK_DELETE, Alt,   self, WM_APP_RUN, run_KillWord);
		esv(VK_DELETE, Ctrl,  self, WM_APP_RUN, run_DeleteRegionOrKillLine);
		esv(VK_INSERT, Ctrl,  self, WM_APP_RUN, run_KillRingSave);
		esv(VK_DELETE, Shift, self, WM_APP_RUN, run_KillRegion);
		esv(VK_INSERT, Shift, self, WM_APP_RUN, run_Yank);
		esv(VK_UP,     Shift, self, WM_APP_RUN, run_ExpandRegionPreviousLine);
		esv(VK_DOWN,   Shift, self, WM_APP_RUN, run_ExpandRegionNextLine);
		esv(VK_LEFT,   Shift, self, WM_APP_RUN, run_ExpandRegionBackwardChar);
		esv(VK_RIGHT,  Shift, self, WM_APP_RUN, run_ExpandRegionForwardChar);
		esv(VK_HOME,   Shift, self, WM_APP_RUN, run_ExpandRegionBeginningOfLine);
		esv(VK_END,    Shift, self, WM_APP_RUN, run_ExpandRegionEndOfLine);
		esv(VK_NEXT,   Shift, self, WM_APP_RUN, run_ExpandRegionScrollUp);
		esv(VK_PRIOR,  Shift, self, WM_APP_RUN, run_ExpandRegionScrollDown);
#undef esv

		// isearch_keymap
		isearch_keymap.SetCharOver(0x00, 0x1f, 0, code,  WM_APP_RUN, run_IsearchExitAndRunChar);
		isearch_keymap.SetCharOver(0x20, 0x7e, 0, code,  WM_APP_RUN, run_IsearchPrintingChar);
		isearch_keymap.SetCharOver(0xa1, 0xdf, 0, code,  WM_APP_RUN, run_IsearchPrintingChar);
#define isc isearch_keymap.SetChar
		isc('G', Ctrl, self, WM_APP_RUN, run_IsearchAbort);
		isc('H', Ctrl, self, WM_APP_RUN, run_IsearchDeleteChar);
		isc('I', Ctrl, code, WM_APP_RUN, run_IsearchPrintingChar);
		isc('J', Ctrl, code, WM_APP_RUN, run_IsearchPrintingChar);
		isc('M', Ctrl, code, WM_APP_RUN, run_IsearchExit);
		isc('R', Ctrl, self, WM_APP_RUN, run_IsearchRepeatBackward);
		isc('S', Ctrl, self, WM_APP_RUN, run_IsearchRepeatForward);
		isc('Y', Ctrl, self, WM_APP_RUN, run_IsearchYank);
		isc('\\',Ctrl, self, WM_APP_RUN, run_ImeToggle);
#undef isc
#define isv isearch_keymap.SetVKey
		BYTE vkey[] = { VK_UP, VK_DOWN, VK_LEFT, VK_RIGHT, VK_HOME, VK_END, VK_NEXT, VK_PRIOR };
		isearch_keymap.SetVKeyOver(vkey, sizeof(vkey), 0, self, WM_APP_RUN, run_IsearchExitAndRunVKey);
		isv(VK_DELETE, Shift, self, WM_APP_RUN, run_IsearchExitAndRunVKey);
		isv(VK_INSERT, Shift, self, WM_APP_RUN, run_IsearchExitAndRunVKey);
		isv(VK_DELETE, 0,     self, WM_APP_RUN, run_DeleteBackwardChar);
#undef isv
    
		// windows
		windows_keymap.SetCharOver(0x20, 0x7e, 0, code,  WM_APP_RUN, run_SelfInsertCommand);
		windows_keymap.SetCharOver(0xa1, 0xdf, 0, code,  WM_APP_RUN, run_SelfInsertCommand);
#define wsc windows_keymap.SetChar
		wsc('A', Ctrl, self, WM_APP_RUN, run_SelectAll);
		wsc('C', Ctrl, self, WM_APP_RUN, run_KillRingSave);
		wsc('F', Ctrl, self, WM_APP_RUN, run_IsearchForward);
		//wsc('H', Ctrl, self, WM_APP_RUN, run_DeleteBackwardChar); // oOH
		wsc('I', Ctrl, code, WM_APP_RUN, run_SelfInsertCommand);
		wsc('J', Ctrl, self, WM_APP_RUN, run_Newline);
		wsc('M', Ctrl, self, WM_APP_RUN, run_Newline);
		wsc('R', Ctrl, self, WM_APP_RUN, run_IsearchBackward);
		wsc('S', Ctrl, self, WM_APP_RUN, run_IsearchForward);
		wsc('V', Ctrl, self, WM_APP_RUN, run_Yank);
		wsc('X', Ctrl, self, WM_APP_RUN, run_KillRegion);
		wsc('Y', Ctrl, self, WM_APP_RUN, run_Undo);
		wsc('Z', Ctrl, self, WM_APP_RUN, run_Undo);
#undef wsc
#define wsv windows_keymap.SetVKey
		wsv(VK_UP,     0,     self, WM_APP_RUN, run_PreviousLine);
		wsv(VK_DOWN,   0,     self, WM_APP_RUN, run_NextLine);
		wsv(VK_LEFT,   0,     self, WM_APP_RUN, run_BackwardChar);
		wsv(VK_RIGHT,  0,     self, WM_APP_RUN, run_ForwardChar);
		wsv(VK_HOME,   0,     self, WM_APP_RUN, run_BeginningOfLine);
		wsv(VK_END,    0,     self, WM_APP_RUN, run_EndOfLine);
		wsv(VK_DELETE, 0,     self, WM_APP_RUN, run_DeleteChar);
		wsv(VK_NEXT,   0,     self, WM_APP_RUN, run_ScrollUp);
		wsv(VK_PRIOR,  0,     self, WM_APP_RUN, run_ScrollDown);
		wsv(VK_BACK,   Alt,   self, WM_APP_RUN, run_SelectAll);
		wsv(VK_LEFT,   Ctrl,  self, WM_APP_RUN, run_BackwardWord);
		wsv(VK_RIGHT,  Ctrl,  self, WM_APP_RUN, run_ForwardWord);
		wsv(VK_HOME,   Ctrl,  self, WM_APP_RUN, run_BeginningOfBuffer);
		wsv(VK_END,    Ctrl,  self, WM_APP_RUN, run_EndOfBuffer);
		wsv(VK_DELETE, Ctrl,  self, WM_APP_RUN, run_DeleteRegionOrKillLine);
		wsv(VK_INSERT, Ctrl,  self, WM_APP_RUN, run_KillRingSave);
		wsv(VK_DELETE, Shift, self, WM_APP_RUN, run_KillRegion);
		wsv(VK_INSERT, Shift, self, WM_APP_RUN, run_Yank);
		wsv(VK_UP,     Shift, self, WM_APP_RUN, run_ExpandRegionPreviousLine);
		wsv(VK_DOWN,   Shift, self, WM_APP_RUN, run_ExpandRegionNextLine);
		wsv(VK_LEFT,   Shift, self, WM_APP_RUN, run_ExpandRegionBackwardChar);
		wsv(VK_RIGHT,  Shift, self, WM_APP_RUN, run_ExpandRegionForwardChar);
		wsv(VK_HOME,   Shift, self, WM_APP_RUN, run_ExpandRegionBeginningOfLine);
		wsv(VK_END,    Shift, self, WM_APP_RUN, run_ExpandRegionEndOfLine);
		wsv(VK_NEXT,   Shift, self, WM_APP_RUN, run_ExpandRegionScrollUp);
		wsv(VK_PRIOR,  Shift, self, WM_APP_RUN, run_ExpandRegionScrollDown);
		wsv(VK_BACK,   Ctrl | Shift, self, WM_APP_RUN, run_BackwardKillWord);
		wsv(VK_DELETE, Ctrl | Shift, self, WM_APP_RUN, run_KillWord);
#undef wsv
	}

	// Isearch ǂ
	int wmAppIsInIsearchMode(void)
	{
		return isearch_mode != 0;
	}

	// Key binding I
	int wmAppSelectKeyBinding(KeyBinding kb)
	{
		if (kb == KeyBindingEmacs || kb == KeyBindingWindows) key_binding = kb;
		else key_binding = KeyBindingEmacs;
		if (current_keymap == &emacs_keymap ||
			current_keymap == &windows_keymap)
			current_keymap = (key_binding == KeyBindingEmacs) ? &emacs_keymap : &windows_keymap;
		return 0;
	}

	// Key bindings map 擾
	int wmAppGetKeyBindingsMap(char which_keymap)
	{
		switch (which_keymap) {
		case 'E': return (int)&emacs_keymap;
		case 'W': return (int)&windows_keymap;
		case 'C': return (int)&ctrl_c_keymap;
		case 'X': return (int)&ctrl_x_keymap;
		case 'I': return (int)&isearch_keymap;
		default:  return 0;
		}
	}

	// WM_KEYDOWN
	int wmKeyDown(int vk_code, long key_data)
	{
		current_keymap->KeyDown(vk_code);
		return 0;
	}

	// WM_SYSKEYDOWN
	int wmSysKeyDown(int vk_code, long key_data)
	{
		if (current_keymap->SysKeyDown(vk_code)) return 0; else return 1;
	}

	// WM_CHAR
	int wmChar(char char_code, long key_data)
	{
		current_keymap->Char(char_code, 0);
		return 0;
	}

	// WM_SYSCHAR
	int wmSysChar(char char_code, long key_data)
	{
		if (current_keymap->SysChar(char_code)) return 0; else return 1;
	}

	// WM_IME_COMPOSITION
	int wmImeComposition(int chDBCS, int flags)
	{
		if (isearch_mode == 0) return 2;    
		if ((flags & GCS_RESULTSTR) == 0) return 1;

		HIMC hIMC = ImmGetContext(Hwnd());
		if (!hIMC) return 1;

		int size = ImmGetCompositionString(hIMC, GCS_RESULTSTR, NULL, 0);
		char *str = (char *)_alloca(size + sizeof(wchar_t));

		ImmGetCompositionString(hIMC, GCS_RESULTSTR, str, size + sizeof(wchar_t));
		ImmReleaseContext(Hwnd(), hIMC);
		str[size] = '\0';
		isearch_string(str);

		return 0;
	}

	// WM_IME_ENDCOMPOSITION
	int wmImeEndComposition(void)
	{
		return isearch_mode == 0;
	}
};


WNDPROC EmacsBinding::original_wndproc;


// Emacs binding Lɂ
void lgcRichEdit::EnableEmacsBinding(lgcStatusbar *statusbar)
{
	if (EmacsBinding::original_wndproc == NULL){
		EmacsBinding::original_wndproc = (WNDPROC)SubclassWindow(Hwnd(), EmacsBindingWndProc);
	} else if (EmacsBinding::original_wndproc == (WNDPROC)GetClassLong(Hwnd(), GCL_WNDPROC)){
		SubclassWindow(Hwnd(), EmacsBindingWndProc);
	}
    
  if (GetWindowLongPtr(Hwnd(), GWLP_USERDATA) == 0)
    SetWindowLongPtr(Hwnd(), GWLP_USERDATA,
		  (LONG_PTR)new EmacsBinding(Hwnd(), statusbar));
}


static LRESULT CALLBACK EmacsBindingWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
#define lgWindowClass EmacsBinding
	if (lgwc) switch (message) {
		lgwmChar;
	case WM_KEYDOWN:
		if (wParam == VK_BACK && !lgwc->wmAppIsInIsearchMode()) break;
		else return lgwc->wmKeyDown((int)wParam, lParam);
	case WM_SYSCHAR:
		if (lgwc->wmSysChar((int)wParam, lParam) == 0) return 0;
		else return DefWindowProc(hwnd, message, wParam, lParam);
	case WM_SYSKEYDOWN:
		if (lgwc->wmSysKeyDown((int)wParam, lParam) == 0) return 0;
		else return DefWindowProc(hwnd, message, wParam, lParam);
	case lgcRichEdit::WM_APP_IsInIsearchMode: // Isearch ǂ
		return lgwc->wmAppIsInIsearchMode();
	case lgcRichEdit::WM_APP_SelectKeyBinding: // Key binding I
		return lgwc->wmAppSelectKeyBinding((lgcRichEdit::KeyBinding)wParam);
	case lgcRichEdit::WM_APP_GetKeyBindingsMap: // Key bindings map 擾
		return lgwc->wmAppGetKeyBindingsMap((char)wParam);
	case WM_APP_RUN:
		return lgwc->wmAppRun(wParam, lParam);
	case WM_IME_COMPOSITION:
		{
			int i = lgwc->wmImeComposition(wParam, lParam);
			if (i == 0) return 0;
			if (i == 1) return DefWindowProc(hwnd, message, wParam, lParam);
			break;
		}
	case WM_IME_ENDCOMPOSITION:
		{
			if (lgwc->wmImeEndComposition() == 0) return DefWindowProc(hwnd, message, wParam, lParam);
			break;
		}
	}
	if (EmacsBinding::original_wndproc == NULL) return DefWindowProc(hwnd, message, wParam, lParam);
	else return CallWindowProc((WNDPROC)EmacsBinding::original_wndproc, hwnd, message, wParam, lParam);
}
