/*
 * LIBGANA - library for GANA
 * Copyright (C) 1996-1997, TAGA Nayuta <nayuta@is.s.u-tokyo.ac.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

///////////////////////////////////////////////////////////////////////////////
// lgcSplitbar.cxx

#include	<crtdbg.h>

#include	"lgcSplitbar.h"
#include	"lgGUIParts.h"
#include	"lgRegistry.h"
#include	"resource.h"
#include	"rWItalk2.h"

LRESULT CALLBACK lgcSplitbarWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

class lgcSplitbar : public lgWindow
{
	HWND children[2];   // lgcSplitbar ̓̎qEBhE HWND
	BOOL is_redraw[2];  // lgcSplitbar ̎qEBhE Redraw 邩ǂ
	BOOL is_horz;       // lgcSplitbar ǂ
	int bar_place;      // lgcSplitbar ̃o[̈ʒu
	int bar_size;       // lgcSplitbar ̃o[̑
	BOOL fixed;         // lgcSplitbar ̃o[𓮂Ȃ
	int move_bar_place; // lgcSplitbar ̃o[ړƂ̃o[̈ʒu
	char*	exchgMessage;

	void set_bar_size(int _bar_size = -1)
	{
		if (_bar_size <= 0) {
			// modified by sept
			//bar_size  = GetSystemMetrics(is_horz ? SM_CYSIZEFRAME : SM_CXSIZEFRAME);
			bar_size  = GetSystemMetrics(is_horz ? SM_CYFIXEDFRAME : SM_CXFIXEDFRAME);
			//bar_size += GetSystemMetrics(is_horz ? SM_CYEDGE      : SM_CXEDGE     );
		} else bar_size = _bar_size;
	}

	// thbOł͂ȂƂɃo[`
	int draw_invert_bar(int _bar_place)
	{
		RECT rc;
		GetClientRect(Hwnd(), &rc);
		if (is_horz) {
			_bar_place = max(0, min(_bar_place, rc.bottom - bar_size));
			rc.top = _bar_place;
			rc.bottom = _bar_place + bar_size;
		} else {
			_bar_place = max(0, min(_bar_place, rc.right - bar_size));
			rc.left = _bar_place;
			rc.right = _bar_place + bar_size;
		}

		HDC hdc = GetWindowDC(Hwnd());
		DrawFocusRect(hdc, &rc);
		ReleaseDC(Hwnd(), hdc);

		return _bar_place;
	}

public:
	lgcSplitbar(HWND _hwnd)
		: lgWindow(_hwnd),
		  fixed(FALSE)
	{
		children[0] = children[1] = NULL;
	}

	int wmCreate(CREATESTRUCT *cs)
	{
		is_horz = !(cs->style & SBS_VERT);
		set_bar_size();

		RECT rc;
		GetClientRect(Hwnd(), &rc);
		if (is_horz) bar_place = (rc.bottom - bar_size) / 2; else bar_place = (rc.right  - bar_size) / 2;
		return 0;
	}

	// o[̈ʒu擾
	int mGetBarPlace(void)
	{
		return bar_place;
	}

	// o[̈ʒuݒ肷
	int mSetBarPlace(int _bar_place)
	{
		HWND hwnd_p = Hwnd();
		while (1) {
			HWND _hwnd_p = GetParent(hwnd_p);
			if (!_hwnd_p) {
				if (IsIconic(hwnd_p)) return -1;
				break;
			}
			hwnd_p = _hwnd_p;
		}
		RECT rc;
		GetClientRect(Hwnd(), &rc);
		bar_place = _bar_place;
		if (is_horz) {
			if (rc.bottom - bar_size <= bar_place) fixed = TRUE;
			if (fixed) bar_place = rc.bottom;
			bar_place = max(0, min(bar_place, rc.bottom - bar_size));
			::MoveWindow(children[0], 0, 0, rc.right, bar_place, TRUE);
			::MoveWindow(children[1], 0, bar_place + bar_size, rc.right, rc.bottom - bar_place - bar_size, TRUE);
			rc.top = bar_place;
			rc.bottom = bar_place + bar_size;
		} else {
			if (rc.right - bar_size <= bar_place) fixed = TRUE;
			if (fixed) bar_place = rc.right;
			bar_place = max(0, min(bar_place, rc.right - bar_size));
			::MoveWindow(children[0], 0, 0, bar_place, rc.bottom, TRUE);
			::MoveWindow(children[1], bar_place + bar_size, 0, rc.right - bar_place - bar_size, rc.bottom, TRUE);
			rc.left = bar_place;
			rc.right = bar_place + bar_size;
		}
		for (int i = 0; i < 2; i++) if (is_redraw[i]) { // ĕ`
			InvalidateRect(children[i], NULL, TRUE);
			UpdateWindow(children[i]);
		}
		InvalidateRect(Hwnd(), &rc, TRUE);
		UpdateWindow(Hwnd());
		return 0;
	}

	// lgcSplitbar ɎqEBhE֘At
	int mSetWindow(int n, HWND child_hwnd)
	{
		children[n % 2] = child_hwnd;
		return 0;
	}
  
	// lgcSplitbar Ɋ֘AtꂽqEBhE擾
	HWND mGetWindow(int n)
	{
		return children[n % 2];
	}

	// o[̈ʒuωɊ֘AtꂽqEBhEĕ`悷邩ǂ
	int mSetRedraw(int n, BOOL _is_redraw)
	{
		is_redraw[n % 2] = _is_redraw;
		return 0;
	}

	// lgcSplitbar  Tooltip ֘At
	LRESULT mSetTooltips(HWND hwnd_tooltip)
	{
		char buf[BUFSIZE];
		::LoadString(lgWindows::lg_hInst(), rStringSplitbarMessage, buf, sizeof(buf));

		TOOLINFO ti;
		ti.cbSize   = sizeof(TOOLINFO);
		ti.uFlags   = TTF_IDISHWND | TTF_SUBCLASS;
		ti.hwnd     = Hwnd();
		ti.uId      = (UINT)Hwnd();
		ti.hinst    = NULL;
		ti.lpszText = buf;

		return SendMessage(hwnd_tooltip, TTM_ADDTOOL, 0, (LPARAM)&ti);
	}

	// CAEgݒ
	int mSetLayout(int layout)
	{
		fixed = FALSE;
		is_horz = (layout & 1); // 1 rbg : ǂ
		if (layout & 2) {       // 2 rbg : ւĂ邩ǂ
			HWND c = children[0];
			children[0] = children[1];
			children[1] = c;
		}
		mSetBarPlace(layout >> 2);
		return 0;
	}

	// CAEg擾
	int mGetLayout(HWND hwnd_child0, HWND hwnd_child1)
	{
		int layout = bar_place << 2;
		layout |= is_horz ? 1 : 0;
		layout |= (hwnd_child0 == children[0] || hwnd_child1 == children[1]) ? 0 : 2;
		return layout;
	}

	// o[̏Ń_uNbNꂽAEBhËʒuւ
	int wmLButtonDblclk(WORD /* keys */, int x, int y)
	{
		// bZ[W{bNXoēꊷĂ悢₢킹
		exchgMessage = LoadString(rStringXchgBuffer);
		if (MessageBox(0, exchgMessage, lgWindows::TheName(), MB_OKCANCEL | MB_DEFBUTTON2 | MB_ICONQUESTION) != IDOK) return 0;
		move_bar_place = -1;
		if (is_horz) {
			HWND c = children[0];
			children[0] = children[1];
			children[1] = c;
		}
		is_horz = !is_horz;
		set_bar_size();
		mSetBarPlace((is_horz ? y : x) - bar_size / 2);
		return 0;
	}

	// o[𓮂̂~߂
	int wmLButtonUp(WORD /* keys */, int /* x */, int /* y */)
	{ return wmButtonUp(0); }
	int wmRButtonUp(WORD /* keys */, int /* x */, int /* y */)
	{ return wmButtonUp(2); }
	int wmButtonUp(int /* n */)
	{
		ReleaseCapture();
		if (0 <= move_bar_place) {
			draw_invert_bar(move_bar_place);
			fixed = FALSE;
			mSetBarPlace(move_bar_place);
		}
		return 0;
	}
  
	// o[𓮂͂߂
	int wmLButtonDown(WORD /* keys */, int /* x */, int /* y */)
	{ return wmButtonDown(0); }
	int wmRButtonDown(WORD /* keys */, int /* x */, int /* y */)
	{ return wmButtonDown(2); }
	int wmButtonDown(int n)
	{
		SetCapture(Hwnd());
		char buf[10];
		lgRegistry::Read(HKEY_CURRENT_USER, "Control Panel\\desktop", "DragfullWindows", buf, sizeof(buf), "0");
		BOOL is_fulldrag = atoi(buf);
		if ((n == 0 && is_fulldrag) || (n == 2 && !is_fulldrag)) move_bar_place = -1; // thbO
		else move_bar_place = draw_invert_bar(bar_place); // thbOł͂Ȃ
		return 0;
	}

	// hbO̓o[ړ
	int wmMouseMove(WORD /* keys */, int x, int y)
	{
		SetCursor(is_horz ? hCursor_SplitbarH : hCursor_SplitbarV);
		if (Hwnd() == GetCapture())
			if (move_bar_place < 0) {
				fixed = FALSE;
				mSetBarPlace((is_horz ? y : x) - bar_size / 2);
			} else {
				draw_invert_bar(move_bar_place);
				move_bar_place = draw_invert_bar((is_horz ? y : x) - bar_size / 2);
			}
		return 0;
	}

	// tH[JX͐eEBhEɂ܂킷
	int wmSetFocus(HWND /* hwnd_lose_focus */)
	{
		SetFocus(GetParent(Hwnd()));
		return 0;
	}

	// EBhETCY̕ύXɍ킹ăo[̈ʒu𒲐߂
	int wmSize(int /* size_type */, int /* width */, int /* height */)
	{
		mSetBarPlace(bar_place);
		return 0;
	}

	// o[`
	int wmPaint(void)
	{
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(Hwnd(), &ps);

		RECT rc;
		GetClientRect(Hwnd(), &rc);

		if (is_horz) {
			rc.top = bar_place;
			rc.bottom = bar_place + bar_size;
		} else {
			rc.left = bar_place;
			rc.right = bar_place + bar_size;
		}
		// modified by sept
		//lgGUIParts::Draw3DFrameInside(hdc, rc, lgGUIParts::FrameUp2, COLOR_3DFACE);
		HBRUSH hbr = CreateSolidBrush(GetSysColor(COLOR_3DFACE));
		FillRect( hdc, &rc, hbr);

		DeleteObject(hbr);

		EndPaint(Hwnd(), &ps);
		return 0;
	}

private:
	static HCURSOR hCursor_SplitbarV; // o[̎̃J[\
	static HCURSOR hCursor_SplitbarH; // o[̎̃J[\

public:

	// lgcSplitbar EBhENXo^
	static ATOM Register(void)
	{
		// r̕KvȂ
		hCursor_SplitbarH = LoadCursor(lgWindows::lg_hInst(), MAKEINTRESOURCE(rCursor_SplitbarH));
		hCursor_SplitbarV = LoadCursor(lgWindows::lg_hInst(), MAKEINTRESOURCE(rCursor_SplitbarV));
		WNDCLASSEX wc;
		wc.cbSize        = sizeof(WNDCLASSEX);
		wc.style         = CS_DBLCLKS;
		wc.lpfnWndProc   = lgcSplitbarWndProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 4;
		wc.hInstance     = lgWindows::hInst();
		wc.hIcon         = NULL;
		wc.hIconSm       = NULL;
		wc.hCursor       = NULL;
		wc.hbrBackground = NULL;
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = WC_lgcSplitbar;
		ATOM rrcex = RegisterClassEx(&wc);
		_ASSERT(rrcex);
		return rrcex;
	}
};


HCURSOR lgcSplitbar::hCursor_SplitbarV;
HCURSOR lgcSplitbar::hCursor_SplitbarH;


LRESULT CALLBACK lgcSplitbarWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
#define lgWindowClass lgcSplitbar
	if (!lgwc) switch (message) {
		lgwmCreate;
	} else switch (message) {
		lgwmDestroy;
		lgwmPaint;
		lgwmSize;
		lgwmSetFocus;
		lgwmLButtonDown;
		lgwmLButtonUp;
		lgwmLButtonDblclk;
		lgwmRButtonDown;
		lgwmRButtonUp;
		lgwmMouseMove;

	  case WM_COMMAND:
	  case WM_NOTIFY:
		  return SendMessage(GetParent(hwnd), message, wParam, lParam);
	  case lgcSplitbar_SETCHILD:
		  return lgwc->mSetWindow(wParam, (HWND)lParam);
	  case lgcSplitbar_GETCHILD: return (LRESULT)lgwc->mGetWindow(wParam);
	  case lgcSplitbar_SETREDRAW: return lgwc->mSetRedraw(wParam, (int)lParam);
	  case lgcSplitbar_SETBARPLACE: return lgwc->mSetBarPlace(lParam);
	  case lgcSplitbar_GETBARPLACE: return lgwc->mGetBarPlace();
	  case lgcSplitbar_SETTOOLTIPS: return lgwc->mSetTooltips((HWND)lParam);
	  case lgcSplitbar_SETLAYOUT: return lgwc->mSetLayout((int)lParam);
	  case lgcSplitbar_GETLAYOUT:
		  return lgwc->mGetLayout((HWND)wParam, (HWND)lParam);
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}


// lgcSplitbar NX̓o^
ATOM Register_lgcSplitbar(void)
{
	return lgcSplitbar::Register();
}
